# -*- coding: utf-8 -*-
# Copyright: (c) 2021, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import base64
import collections
import contextlib
import os
import queue
import socket
import subprocess
import threading
import typing
import uuid
from xml.etree import ElementTree

import pytest
from xmldiff import main as _xmldiff

if os.name == "nt":
    import win32api
    import win32event
    import win32file
    import win32pipe
    import winerror

import psrpcore

BUFFER_SIZE = 327681

# Contains control characters, non-ascii chars, and chars that are surrogate pairs in UTF-16
COMPLEX_STRING = "treble clef\n _x0000_ _X0000_ %s café \uD83C" % b"\xF0\x9D\x84\x9E".decode("utf-8")
COMPLEX_ENCODED_STRING = "treble clef_x000A_ _x005F_x0000_ _x005F_X0000_ _xD834__xDD1E_ café _xD83C_"

T = typing.TypeVar("T", psrpcore.ClientRunspacePool, psrpcore.ServerRunspacePool)

OutOfProcPacket = collections.namedtuple("OutOfProcPacket", ["action", "ps_guid", "data"])


def which(program: str) -> typing.Optional[str]:
    for path in os.environ.get("PATH", "").split(os.pathsep):
        exe = os.path.join(path, program)
        if os.path.isfile(exe) and os.access(exe, os.X_OK):
            return exe

    return


PWSH_PATH = which("pwsh.exe" if os.name == "nt" else "pwsh")


class FakeCryptoProvider(psrpcore.types.PSCryptoProvider):
    def decrypt(self, value: bytes) -> bytes:
        return value

    def encrypt(self, value: bytes) -> bytes:
        return value

    def register_key(self, key: bytes) -> None:
        pass


class OutOfProcTransport(typing.Generic[T]):
    def __init__(self, runspace: T) -> None:
        self.runspace = runspace
        self._incoming: queue.Queue[typing.Union[Exception, OutOfProcPacket]] = queue.Queue()
        self._listen_task = threading.Thread(target=self._read_task)
        self._wait = threading.Condition()
        self._wait_set = set()

    def __enter__(self) -> "OutOfProcTransport":
        self._open()
        self._listen_task.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self._close()
        self._listen_task.join()

    def next_payload(self) -> OutOfProcPacket:
        # No long running tests, anything taking more than 60 seconds is a failure
        payload = self._incoming.get(timeout=60)
        if isinstance(payload, Exception):
            raise payload

        return payload

    def next_event(self) -> psrpcore.PSRPEvent:
        while True:
            event = self.runspace.next_event()
            if event:
                return event

            payload = self.next_payload()
            if payload.action != "Data":
                continue

            self.runspace.receive_data(payload.data)

    def close(
        self,
        pipeline_id: typing.Optional[uuid.UUID] = None,
    ) -> None:
        with self._wait_ack("Close", pipeline_id):
            self._send(ps_guid_packet("Close", pipeline_id))

    def close_ack(
        self,
        pipeline_id: typing.Optional[uuid.UUID] = None,
    ) -> None:
        self._send(ps_guid_packet("CloseAck", pipeline_id))

    def command(
        self,
        pipeline_id: typing.Optional[uuid.UUID] = None,
    ) -> None:
        with self._wait_ack("Command", pipeline_id):
            self._send(ps_guid_packet("Command", pipeline_id))

    def command_ack(
        self,
        pipeline_id: typing.Optional[uuid.UUID] = None,
    ) -> None:
        self._send(ps_guid_packet("CommandAck", pipeline_id))

    def data(
        self,
        wait_ack: bool = True,
    ) -> bool:
        data = self.runspace.data_to_send()
        if not data:
            return False

        with self._wait_ack("Data" if wait_ack else None, data.pipeline_id):
            self._send(ps_data_packet(*data))

        return True

    def data_ack(
        self,
        pipeline_id: typing.Optional[uuid.UUID] = None,
    ) -> None:
        self._send(ps_guid_packet("DataAck", pipeline_id))

    def signal(
        self,
        pipeline_id: typing.Optional[uuid.UUID] = None,
    ) -> None:
        with self._wait_ack("Signal", pipeline_id):
            self._send(ps_guid_packet("Signal", pipeline_id))

    def signal_ack(
        self,
        pipeline_id: typing.Optional[uuid.UUID] = None,
    ) -> None:
        self._send(ps_guid_packet("SignalAck", pipeline_id))

    def _read_task(self):
        try:
            self._read()

        except Exception as e:
            self._incoming.put(e)

        finally:
            with self._wait:
                for key in list(self._wait_set):
                    self._wait_set.remove(key)
                self._wait.notify_all()

    def _read(self) -> None:
        buffer = bytearray()
        while True:
            try:
                end_idx = buffer.index(b"\n")
            except ValueError:
                # Don't have enough data - wait for more to arrive.
                read_data = self._recv()
                if not read_data:
                    break

                buffer += read_data
                continue

            raw_element = bytes(buffer[:end_idx])
            buffer = buffer[end_idx + 1 :]

            element = ElementTree.fromstring(raw_element)
            ps_guid = uuid.UUID(element.attrib["PSGuid"].upper())
            if ps_guid == uuid.UUID(int=0):
                ps_guid = None

            if element.tag == "Data":
                psrp_data = base64.b64decode(element.text) if element.text else b""
                stream_type = (
                    psrpcore.StreamType.prompt_response
                    if element.attrib.get("Stream", "") == "PromptResponse"
                    else psrpcore.StreamType.default
                )
                payload = psrpcore.PSRPPayload(psrp_data, stream_type, ps_guid)
                self._incoming.put(OutOfProcPacket("Data", ps_guid, payload))

            elif element.tag.endswith("Ack"):
                pipeline = str(ps_guid) if ps_guid else ""
                with self._wait:
                    self._wait_set.remove(f"{element.tag}:{pipeline.upper()}")
                    self._wait.notify_all()

            else:
                self._incoming.put(OutOfProcPacket(element.tag, ps_guid, None))

    def _open(self) -> None:
        raise NotImplementedError()

    def _close(self) -> None:
        raise NotImplementedError()

    def _recv(self) -> typing.Optional[bytes]:
        raise NotImplementedError()

    def _send(self, data: bytes) -> None:
        raise NotImplementedError()

    @contextlib.contextmanager
    def _wait_ack(
        self,
        action: typing.Optional[str],
        pipeline_id: typing.Optional[uuid.UUID] = None,
    ):
        if not action:
            yield
            return

        pipeline = str(pipeline_id) if pipeline_id else ""
        key = f"{action}Ack:{pipeline.upper()}"
        with self._wait:
            self._wait_set.add(key)
            yield
            self._wait.wait_for(lambda: key not in self._wait_set)


class ClientTransport(OutOfProcTransport[psrpcore.ClientRunspacePool]):
    def __init__(self, runspace: psrpcore.ClientRunspacePool, executable: str) -> None:
        super().__init__(runspace)
        self._executable = executable
        self._process = None

    def _open(self) -> None:
        pipe = subprocess.PIPE
        self._process = subprocess.Popen([self._executable, "-NoLogo", "-s"], stdin=pipe, stdout=pipe, stderr=pipe)

    def _close(self) -> None:
        if self._process.poll() is None:
            self._process.kill()
            self._process.wait()

    def _recv(self) -> typing.Optional[bytes]:
        stdout = self._process.stdout.readline()
        if not stdout:
            stdout, stderr = self._process.communicate()
            if stderr:
                raise Exception(stderr.decode())

        return stdout

    def _send(self, data: bytes) -> None:
        self._process.stdin.write(data)
        self._process.stdin.flush()


class ServerTransport(OutOfProcTransport[psrpcore.ServerRunspacePool]):
    def __init__(self, runspace: psrpcore.ServerRunspacePool, pipe_name: str) -> None:
        super().__init__(runspace)
        self.pipe_name = pipe_name

    def data(
        self,
        wait_ack: bool = False,
    ) -> bool:
        return super().data(wait_ack=wait_ack)


class UnixDomainSocket(ServerTransport):
    def __init__(self, runspace: psrpcore.ServerRunspacePool, pipe_name: str) -> None:
        super().__init__(runspace, pipe_name)
        self._sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self._conn: typing.Optional[socket.socket] = None
        self._sock_lock = threading.Lock()

    def _open(self) -> None:
        self._sock.bind(self.pipe_name)
        self._sock.listen(1)

    def _close(self) -> None:
        if self._conn:
            self._conn.shutdown(socket.SHUT_RDWR)
            self._conn.close()

        self._conn = None

        self._sock.shutdown(socket.SHUT_RDWR)
        self._sock.close()

    def _recv(self) -> typing.Optional[bytes]:
        return self._get_sock().recv(BUFFER_SIZE)

    def _send(self, data: bytes) -> None:
        self._get_sock().sendall(data)

    def _get_sock(self) -> socket.socket:
        with self._sock_lock:
            if not self._conn:
                self._conn = self._sock.accept()[0]

            return self._conn


class NamedPipe(ServerTransport):
    def __init__(self, runspace: psrpcore.ServerRunspacePool, pipe_name: str) -> None:
        super().__init__(runspace, pipe_name)
        self._pipe = win32pipe.CreateNamedPipe(
            "\\\\.\\pipe\\" + pipe_name,
            win32pipe.PIPE_ACCESS_DUPLEX | win32file.FILE_FLAG_OVERLAPPED,
            win32pipe.PIPE_TYPE_MESSAGE | win32pipe.PIPE_READMODE_MESSAGE | win32pipe.PIPE_WAIT,
            1,
            BUFFER_SIZE,
            BUFFER_SIZE,
            0,
            None,
        )
        self._connected = False
        self._end_event = win32event.CreateEvent(None, True, 0, None)
        self._write_requested = win32event.CreateEvent(None, True, 0, None)
        self._pipe_lock = threading.Lock()
        self._write_lock = threading.Lock()

    def _open(self) -> None:
        pass

    def _close(self) -> None:
        win32event.SetEvent(self._end_event)
        with self._pipe_lock:
            if self._connected:
                win32pipe.DisconnectNamedPipe(self._pipe)
                self._connected = False

            if self._pipe:
                win32file.CloseHandle(self._pipe)
                self._pipe = None

    def _recv(self) -> typing.Optional[bytes]:
        buffer = win32file.AllocateReadBuffer(BUFFER_SIZE)
        overlapped = win32file.OVERLAPPED()
        overlapped.hEvent = win32event.CreateEvent(None, True, 0, None)

        while True:
            with self._write_lock:
                pass

            with self._pipe_lock:
                self._connect()
                res = win32file.ReadFile(self._pipe, buffer, overlapped)[0]

                if res == winerror.ERROR_SUCCESS:
                    bytes_read = win32file.GetOverlappedResult(self._pipe, overlapped, True)
                    return bytes(buffer[:bytes_read])

                elif res != winerror.ERROR_IO_PENDING:
                    msg = win32api.FormatMessage(res)
                    raise Exception(f"Named pipe read failed 0x{res:08X} - {msg}")

                wait_idx = win32event.WaitForMultipleObjects(
                    (overlapped.hEvent, self._end_event, self._write_requested),
                    False,
                    win32event.INFINITE,
                )

                if wait_idx != win32event.WAIT_OBJECT_0:
                    win32file.CancelIo(self._pipe)

                try:
                    bytes_read = win32file.GetOverlappedResult(self._pipe, overlapped, True)
                except win32file.error as err:
                    if err.winerror != winerror.ERROR_OPERATION_ABORTED:
                        raise
                    bytes_read = 0

                if bytes_read:
                    data = bytes(buffer[:bytes_read])
                    return data

                elif wait_idx == win32event.WAIT_OBJECT_0 + 1:
                    return

    def _send(self, data: bytes) -> None:
        with self._write_lock:
            win32event.SetEvent(self._write_requested)

            with self._pipe_lock:
                buffer = bytearray(data)
                offset = 0
                self._connect()

                while offset < len(data):
                    res, bytes_written = win32file.WriteFile(self._pipe, buffer[offset:])
                    if res != winerror.ERROR_SUCCESS:
                        msg = win32api.FormatMessage(res)
                        raise Exception(f"Named pipe write failed 0x{res:08X} - {msg}")

                    offset += bytes_written
                    win32file.FlushFileBuffers(self._pipe)

                win32event.ResetEvent(self._write_requested)

    def _connect(self):
        if not self._connected:
            overlapped = win32file.OVERLAPPED()
            overlapped.hEvent = win32event.CreateEvent(None, True, 0, None)

            res = win32pipe.ConnectNamedPipe(self._pipe, overlapped)

            if res == winerror.ERROR_IO_PENDING:
                wait_idx = win32event.WaitForMultipleObjects(
                    (overlapped.hEvent, self._end_event),
                    False,
                    win32event.INFINITE,
                )
                if wait_idx != win32event.WAIT_OBJECT_0:
                    raise Exception("Failed while waiting for client connection")

            elif res != winerror.ERROR_SUCCESS:
                msg = win32api.FormatMessage(res)
                raise Exception(f"Named pipe connect failed 0x{res:08X} - {msg}")

            self._connected = True


def get_runspace_pair(
    min_runspaces: int = 1, max_runspaces: int = 1
) -> typing.Tuple[psrpcore.ClientRunspacePool, psrpcore.ServerRunspacePool]:
    client = psrpcore.ClientRunspacePool(min_runspaces=min_runspaces, max_runspaces=max_runspaces)
    server = psrpcore.ServerRunspacePool()

    client.open()
    server.receive_data(client.data_to_send())
    server.next_event()
    server.next_event()
    client.receive_data(server.data_to_send())
    client.next_event()
    client.next_event()
    client.next_event()

    return client, server


def assert_xml_diff(actual: str, expected: str):
    # We don't care that the XML text is the exact same but rather if they represent the same object. Python versions
    # vary on how they order attributes of an element whereas xmldiff doesn't care.
    diff = _xmldiff.diff_texts(actual, expected)
    if len(diff) != 0:
        # The assertion for diff_texts isn't pretty and it's easier to see what the diff is by comparing the text.
        assert actual == expected


def serialize(value: typing.Any, **kwargs: typing.Any) -> ElementTree.Element:
    return psrpcore.types.serialize(value, FakeCryptoProvider(), **kwargs)


def deserialize(value: ElementTree.Element, **kwargs: typing.Any):
    return psrpcore.types.deserialize(value, FakeCryptoProvider(), **kwargs)


def ps_data_packet(
    data: bytes,
    stream_type: psrpcore.StreamType = psrpcore.StreamType.default,
    ps_guid: typing.Optional[uuid.UUID] = None,
) -> bytes:
    """Data packet for PSRP fragments.

    This creates a data packet that is used to encode PSRP fragments when
    sending to the server.

    Args:
        data: The PSRP fragments to encode.
        stream_type: The stream type to target, Default or PromptResponse.
        ps_guid: Set to `None` or a 0'd UUID to target the RunspacePool,
            otherwise this should be the pipeline UUID.

    Returns:
        bytes: The encoded data XML packet.
    """
    ps_guid = ps_guid or uuid.UUID(int=0)
    stream_name = b"Default" if stream_type == psrpcore.StreamType.default else b"PromptResponse"
    return b"<Data Stream='%s' PSGuid='%s'>%s</Data>\n" % (
        stream_name,
        str(ps_guid).lower().encode(),
        base64.b64encode(data),
    )


def ps_guid_packet(
    element: str,
    ps_guid: typing.Optional[uuid.UUID] = None,
) -> bytes:
    """Common PSGuid packet for PSRP message.

    This creates a PSGuid packet that is used to signal events and stages in
    the PSRP exchange. Unlike the data packet this does not contain any PSRP
    fragments.

    Args:
        element: The element type, can be DataAck, Command, CommandAck, Close,
            CloseAck, Signal, and SignalAck.
        ps_guid: Set to `None` or a 0'd UUID to target the RunspacePool,
            otherwise this should be the pipeline UUID.

    Returns:
        bytes: The encoded PSGuid packet.
    """
    ps_guid = ps_guid or uuid.UUID(int=0)
    return b"<%s PSGuid='%s' />\n" % (element.encode(), str(ps_guid).lower().encode())


def run_pipeline(
    client_pwsh: ClientTransport,
    script: str,
    host: typing.Optional[psrpcore.types.HostInfo] = None,
) -> typing.List[psrpcore.PSRPEvent]:
    ps = psrpcore.ClientPowerShell(client_pwsh.runspace, host=host)
    ps.add_script(script)
    ps.start()
    client_pwsh.command(ps.pipeline_id)
    client_pwsh.data()
    events = []
    while ps.state == psrpcore.types.PSInvocationState.Running:
        events.append(client_pwsh.next_event())
    ps.close()
    client_pwsh.close(ps.pipeline_id)

    return events


@pytest.fixture(scope="function")
def client_pwsh():
    """Creates an unopened Runspace Pool against a pwsh process."""
    if not PWSH_PATH:
        pytest.skip("Integration test requires pwsh")

    runspace = psrpcore.ClientRunspacePool()
    with ClientTransport(runspace, PWSH_PATH) as conn:
        yield conn


@pytest.fixture(scope="function")
def client_opened_pwsh():
    """Creates an Opened Runspace Pool against a pwsh process."""
    if not PWSH_PATH:
        pytest.skip("Integration test requires pwsh")

    host = psrpcore.types.HostInfo(
        IsHostNull=False,
        IsHostUINull=False,
        IsHostRawUINull=False,
        UseRunspaceHost=False,
        HostDefaultData=psrpcore.types.HostDefaultData(
            ForegroundColor=psrpcore.types.ConsoleColor.Blue,
            BackgroundColor=psrpcore.types.ConsoleColor.Red,
            CursorPosition=psrpcore.types.Coordinates(X=10, Y=20),
            WindowPosition=psrpcore.types.Coordinates(X=30, Y=40),
            CursorSize=5,
            BufferSize=psrpcore.types.Size(Width=60, Height=120),
            WindowSize=psrpcore.types.Size(Width=60, Height=120),
            MaxWindowSize=psrpcore.types.Size(Width=60, Height=120),
            MaxPhysicalWindowSize=psrpcore.types.Size(Width=60, Height=120),
            WindowTitle="My Window",
        ),
    )

    runspace = psrpcore.ClientRunspacePool(host=host)
    with ClientTransport(runspace, PWSH_PATH) as pwsh:
        pwsh.runspace.open()
        pwsh.data()
        while pwsh.runspace.state == psrpcore.types.RunspacePoolState.Opening:
            pwsh.next_event()

        yield pwsh


@pytest.fixture(scope="function")
def server_pwsh(tmp_path):
    """Creates a pipe PSRP server that can be called in pwsh."""
    if not PWSH_PATH:
        pytest.skip("Integration test requires pwsh")

    pipe_name = f"psrpcore-{str(uuid.uuid4()).upper()}"

    runspace = psrpcore.ServerRunspacePool()
    try:
        if os.name == "nt":
            transport = NamedPipe(runspace, pipe_name)

        else:
            pipe_name = str(tmp_path / pipe_name)
            transport = UnixDomainSocket(runspace, pipe_name)

        with transport:
            yield transport

    finally:
        if os.name != "nt":
            try:
                os.unlink(pipe_name)
            except FileNotFoundError:
                pass



# -*- coding: utf-8 -*-
# Copyright: (c) 2021, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import os
import queue
import shutil
import threading
import typing

import pytest

import psrpcore

from .conftest import COMPLEX_STRING, ClientTransport, ServerTransport


def open_runspace(server_pwsh: ServerTransport) -> None:
    while server_pwsh.runspace.state in [
        psrpcore.types.RunspacePoolState.BeforeOpen,
        psrpcore.types.RunspacePoolState.Opening,
    ]:
        data = server_pwsh.next_payload()
        server_pwsh.runspace.receive_data(data.data)
        while True:
            if not server_pwsh.runspace.next_event():
                break

    while server_pwsh.data():
        pass
    server_pwsh.data_ack()


class BackgroundPipeline(threading.Thread):
    def __init__(
        self,
        client: ClientTransport,
        cmd: str,
        **kwargs: typing.Any,
    ) -> None:
        self.client = client
        self.events = queue.Queue()
        self.pipeline = psrpcore.ClientPowerShell(client.runspace)
        self.pipeline.add_script(cmd)
        if kwargs:
            self.pipeline.add_parameters(**kwargs)

        super().__init__(daemon=True)

    @property
    def remaining_events(self) -> typing.List[psrpcore.PSRPEvent]:
        event_list = []
        while True:
            try:
                event_list.append(self.events.get_nowait())
            except queue.Empty:
                break

        return event_list

    def run(self) -> None:
        self.pipeline.start()
        self.client.command(self.pipeline.pipeline_id)
        self.client.data()

        while self.pipeline.state == psrpcore.types.PSInvocationState.Running:
            e = self.client.next_event()
            if isinstance(e, psrpcore.PipelineOutputEvent) and isinstance(e.data, psrpcore.types.PSSecureString):
                self.client.runspace.exchange_key()
                self.client.data()
                self.client.next_event()

            self.events.put(e)

    def __enter__(self) -> "BackgroundPipeline":
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        # Not ideal but ending the test closes the process and thus cleans up the resource. Because this pipeline is
        # probably connected to the test server trying to actually stop/close if it's still running will hang so just
        # rely on the client closing the entire process.
        if self.pipeline.state != psrpcore.types.PSInvocationState.Running:
            self.pipeline.close()
            self.client.close(self.pipeline.pipeline_id)
            self.join()


def test_open_and_close(server_pwsh: ServerTransport, client_opened_pwsh: ClientTransport):
    cmd = """[CmdletBinding()]
    param ([String]$Name)

    $connInfo = [System.Management.Automation.Runspaces.NamedPipeConnectionInfo]::new($Name)
    $runspacePool = [RunspaceFactory]::CreateRunspacePool(1, 1, $connInfo)
    $runspacePool.Open()
    $runspacePool.Dispose()
    """

    runspace = server_pwsh.runspace
    assert runspace.state == psrpcore.types.RunspacePoolState.BeforeOpen

    with BackgroundPipeline(client_opened_pwsh, cmd, Name=server_pwsh.pipe_name):
        connect = server_pwsh.next_payload()
        assert connect.action == "Data"
        assert connect.ps_guid is None

        runspace.receive_data(connect.data)
        assert runspace.state == psrpcore.types.RunspacePoolState.Opening

        session_cap = runspace.next_event()
        assert isinstance(session_cap, psrpcore.SessionCapabilityEvent)
        assert session_cap.ps_version == runspace.their_capability.PSVersion
        assert session_cap.protocol_version == runspace.their_capability.protocolversion
        assert session_cap.serialization_version == runspace.their_capability.SerializationVersion
        assert runspace.state == psrpcore.types.RunspacePoolState.Opening

        init_runspace = runspace.next_event()
        assert isinstance(init_runspace, psrpcore.InitRunspacePoolEvent)
        assert init_runspace.max_runspaces == 1
        assert init_runspace.min_runspaces == 1
        assert init_runspace.ps_thread_options == psrpcore.types.PSThreadOptions.Default
        assert init_runspace.apartment_state == psrpcore.types.ApartmentState.Unknown
        assert init_runspace.host_info.IsHostNull is True
        assert init_runspace.host_info.IsHostUINull is True
        assert init_runspace.host_info.IsHostRawUINull is True
        assert init_runspace.host_info.UseRunspaceHost is False
        assert init_runspace.host_info.HostDefaultData is None
        assert "PSVersionTable" in init_runspace.application_arguments

        assert runspace.state == psrpcore.types.RunspacePoolState.Opened
        assert runspace.max_runspaces == 1
        assert runspace.min_runspaces == 1
        assert runspace.thread_options == init_runspace.ps_thread_options
        assert runspace.apartment_state == init_runspace.apartment_state
        assert runspace.host.IsHostNull is True
        assert runspace.host.IsHostUINull is True
        assert runspace.host.IsHostRawUINull is True
        assert runspace.host.UseRunspaceHost is False
        assert runspace.host.HostDefaultData is None
        assert runspace.application_arguments == init_runspace.application_arguments
        assert runspace.next_event() is None

        server_pwsh.data()
        server_pwsh.data_ack()

        close = server_pwsh.next_payload()
        assert close.action == "Close"
        assert close.ps_guid is None
        server_pwsh.close_ack()


def test_rp_app_args(server_pwsh: ServerTransport, client_opened_pwsh: ClientTransport):
    cmd = """[CmdletBinding()]
    param (
        [String]$Name,
        [Hashtable]$Arguments
    )

    $connInfo = [System.Management.Automation.Runspaces.NamedPipeConnectionInfo]::new($Name)
    $runspacePool = [RunspaceFactory]::CreateRunspacePool(1, 1, $connInfo, $null, $null, $Arguments)
    $runspacePool.Open()
    $runspacePool.Dispose()
    """

    runspace = server_pwsh.runspace
    assert runspace.state == psrpcore.types.RunspacePoolState.BeforeOpen

    with BackgroundPipeline(
        client_opened_pwsh,
        cmd,
        Name=server_pwsh.pipe_name,
        Arguments={
            "test1": 1,
            "test2": "2",
        },
    ):
        connect = server_pwsh.next_payload()
        assert connect.action == "Data"
        assert connect.ps_guid is None

        runspace.receive_data(connect.data)
        session_cap = runspace.next_event()
        assert isinstance(session_cap, psrpcore.SessionCapabilityEvent)

        init_runspace = runspace.next_event()
        assert isinstance(init_runspace, psrpcore.InitRunspacePoolEvent)
        assert init_runspace.application_arguments["test1"] == 1
        assert init_runspace.application_arguments["test2"] == "2"

        assert runspace.state == psrpcore.types.RunspacePoolState.Opened
        assert runspace.application_arguments == init_runspace.application_arguments
        assert runspace.next_event() is None

        server_pwsh.data()
        server_pwsh.data_ack()
        server_pwsh.next_payload()
        server_pwsh.close_ack()


def test_pipe_output(server_pwsh: ServerTransport, client_opened_pwsh: ClientTransport):
    cmd = """[CmdletBinding()]
    param ([String]$Name, [String]$Script)

    $connInfo = [System.Management.Automation.Runspaces.NamedPipeConnectionInfo]::new($Name)
    $runspacePool = [RunspaceFactory]::CreateRunspacePool(1, 1, $connInfo)
    $runspacePool.Open()
    try {
        $ps = [PowerShell]::Create()
        $ps.RunspacePool = $runspacePool
        [void]$ps.AddScript($Script)
        $ps.Invoke()
    }
    finally {
        $runspacePool.Dispose()
    }
    """

    with BackgroundPipeline(
        client_opened_pwsh,
        cmd,
        Name=server_pwsh.pipe_name,
        Script="my script",
    ) as ps:

        runspace = server_pwsh.runspace
        open_runspace(server_pwsh)

        command = server_pwsh.next_payload()
        assert command.action == "Command"
        assert command.ps_guid is not None
        server_pwsh.command_ack(command.ps_guid)

        s_ps = psrpcore.ServerPipeline(runspace, command.ps_guid)
        create_pipe = server_pwsh.next_event()
        assert isinstance(create_pipe, psrpcore.CreatePipelineEvent)
        assert isinstance(create_pipe.pipeline, psrpcore.PowerShell)
        assert create_pipe.pipeline_id == s_ps.pipeline_id
        assert create_pipe.pipeline.no_input is True
        assert len(create_pipe.pipeline.commands) == 1
        assert create_pipe.pipeline.commands[0].command_text == "my script"

        s_ps.start()
        s_ps.write_output(COMPLEX_STRING)
        s_ps.write_output(psrpcore.types.PSSecureString("secret"))
        s_ps.complete()
        server_pwsh.data()

        pub_key = server_pwsh.next_event()
        assert isinstance(pub_key, psrpcore.PublicKeyEvent)
        server_pwsh.data()

        close = server_pwsh.next_payload()
        assert close.action == "Close"
        assert close.ps_guid == s_ps.pipeline_id
        server_pwsh.close_ack(close.ps_guid)

        out = ps.events.get()
        assert isinstance(out, psrpcore.PipelineOutputEvent)
        assert out.data == COMPLEX_STRING

        out = ps.events.get()
        assert isinstance(out, psrpcore.PipelineOutputEvent)
        assert isinstance(out.data, psrpcore.types.PSSecureString)
        assert out.data.decrypt() == "secret"

        close = server_pwsh.next_payload()
        assert close.action == "Close"
        assert close.ps_guid is None
        server_pwsh.close_ack()


def test_merge_unclaimed(server_pwsh: ServerTransport, client_opened_pwsh: ClientTransport):
    cmd = """[CmdletBinding()]
    param ([String]$Name)

    $connInfo = [System.Management.Automation.Runspaces.NamedPipeConnectionInfo]::new($Name)
    $runspacePool = [RunspaceFactory]::CreateRunspacePool(1, 1, $connInfo)
    $runspacePool.Open()
    try {
        $ps = [PowerShell]::Create()
        $ps.RunspacePool = $runspacePool
        [void]$ps.AddCommand("Write-Error").AddParameter("Message", "Error").AddCommand("Write-Output")
        $ps.Commands.Commands[1].MergeUnclaimedPreviousCommandResults = 'Output, Error'
        $ps.Invoke()
    }
    finally {
        $runspacePool.Dispose()
    }
    """

    with BackgroundPipeline(
        client_opened_pwsh,
        cmd,
        Name=server_pwsh.pipe_name,
    ):

        runspace = server_pwsh.runspace
        open_runspace(server_pwsh)

        command = server_pwsh.next_payload()
        assert command.action == "Command"
        server_pwsh.command_ack(command.ps_guid)

        s_ps = psrpcore.ServerPipeline(runspace, command.ps_guid)
        create_pipe = server_pwsh.next_event()
        assert isinstance(create_pipe, psrpcore.CreatePipelineEvent)
        assert isinstance(create_pipe.pipeline, psrpcore.PowerShell)
        assert create_pipe.pipeline_id == s_ps.pipeline_id
        assert len(create_pipe.pipeline.commands) == 2
        assert create_pipe.pipeline.no_input is True

        cmd1 = create_pipe.pipeline.commands[0]
        assert cmd1.command_text == "Write-Error"
        assert cmd1.parameters == [("Message", "Error")]
        assert cmd1.end_of_statement is False
        assert cmd1.is_script is False
        assert cmd1.merge_unclaimed is False
        assert cmd1.merge_my == psrpcore.types.PipelineResultTypes.none
        assert cmd1.merge_to == psrpcore.types.PipelineResultTypes.none
        assert cmd1.merge_error == psrpcore.types.PipelineResultTypes.none
        assert cmd1.merge_warning == psrpcore.types.PipelineResultTypes.none
        assert cmd1.merge_verbose == psrpcore.types.PipelineResultTypes.none
        assert cmd1.merge_debug == psrpcore.types.PipelineResultTypes.none
        assert cmd1.merge_information == psrpcore.types.PipelineResultTypes.none

        cmd2 = create_pipe.pipeline.commands[1]
        assert cmd2.command_text == "Write-Output"
        assert cmd2.parameters == []
        assert cmd2.end_of_statement is True
        assert cmd2.is_script is False
        assert cmd2.merge_unclaimed is True
        assert cmd2.merge_my == psrpcore.types.PipelineResultTypes.none
        assert cmd2.merge_to == psrpcore.types.PipelineResultTypes.none
        assert cmd2.merge_error == psrpcore.types.PipelineResultTypes.none
        assert cmd2.merge_warning == psrpcore.types.PipelineResultTypes.none
        assert cmd2.merge_verbose == psrpcore.types.PipelineResultTypes.none
        assert cmd2.merge_debug == psrpcore.types.PipelineResultTypes.none
        assert cmd2.merge_information == psrpcore.types.PipelineResultTypes.none

        s_ps.start()
        s_ps.write_output("Error")
        s_ps.complete()
        server_pwsh.data()

        close = server_pwsh.next_payload()
        server_pwsh.close_ack(close.ps_guid)
        server_pwsh.next_payload()
        server_pwsh.close_ack(None)


def test_merge_pipe_out(server_pwsh: ServerTransport, client_opened_pwsh: ClientTransport):
    cmd = """[CmdletBinding()]
    param ([String]$Name)

    $connInfo = [System.Management.Automation.Runspaces.NamedPipeConnectionInfo]::new($Name)
    $runspacePool = [RunspaceFactory]::CreateRunspacePool(1, 1, $connInfo)
    $runspacePool.Open()
    try {
        $ps = [PowerShell]::Create()
        $ps.RunspacePool = $runspacePool
        [void]$ps.AddScript(@'
$VerbosePreference = 'Continue'
$DebugPreference = 'Continue'
$WarningPreference = 'Continue'

"output"
Write-Error -Message "error"
Write-Verbose -Message "verbose"
Write-Debug -Message "debug"
Write-Warning -Message "warning"
Write-Information -MessageData "information"
'@)
        $ps.Commands.Commands[0].MergeMyResults("All", "Output")
        $ps.Invoke()
    }
    finally {
        $runspacePool.Dispose()
    }
    """

    with BackgroundPipeline(
        client_opened_pwsh,
        cmd,
        Name=server_pwsh.pipe_name,
    ):

        runspace = server_pwsh.runspace
        open_runspace(server_pwsh)

        command = server_pwsh.next_payload()
        assert command.action == "Command"
        server_pwsh.command_ack(command.ps_guid)

        s_ps = psrpcore.ServerPipeline(runspace, command.ps_guid)
        create_pipe = server_pwsh.next_event()
        assert isinstance(create_pipe, psrpcore.CreatePipelineEvent)
        assert isinstance(create_pipe.pipeline, psrpcore.PowerShell)
        assert create_pipe.pipeline_id == s_ps.pipeline_id
        assert len(create_pipe.pipeline.commands) == 1
        assert create_pipe.pipeline.no_input is True

        cmd1 = create_pipe.pipeline.commands[0]
        assert cmd1.parameters == []
        assert cmd1.end_of_statement is True
        assert cmd1.is_script is True
        assert cmd1.merge_unclaimed is False
        assert cmd1.merge_my == psrpcore.types.PipelineResultTypes.none
        assert cmd1.merge_to == psrpcore.types.PipelineResultTypes.none
        assert cmd1.merge_error == psrpcore.types.PipelineResultTypes.Output
        assert cmd1.merge_warning == psrpcore.types.PipelineResultTypes.Output
        assert cmd1.merge_verbose == psrpcore.types.PipelineResultTypes.Output
        assert cmd1.merge_debug == psrpcore.types.PipelineResultTypes.Output
        assert cmd1.merge_information == psrpcore.types.PipelineResultTypes.Output

        s_ps.start()
        s_ps.complete()
        server_pwsh.data()

        close = server_pwsh.next_payload()
        server_pwsh.close_ack(close.ps_guid)
        server_pwsh.next_payload()
        server_pwsh.close_ack(None)


def test_progress_record(server_pwsh: ServerTransport, client_opened_pwsh: ClientTransport):
    cmd = """[CmdletBinding()]
    param ([String]$Name)

    $connInfo = [System.Management.Automation.Runspaces.NamedPipeConnectionInfo]::new($Name)
    $runspacePool = [RunspaceFactory]::CreateRunspacePool(1, 1, $connInfo)
    $runspacePool.Open()
    try {
        $ps = [PowerShell]::Create()
        $ps.RunspacePool = $runspacePool
        [void]$ps.AddCommand("Write-Progress").AddParameters([Ordered]@{
            Activity = "act"
            Status = "status"
            Id = 10
            PercentComplete = 34
            SecondsRemaining = 102
            CurrentOperation = "currentOp"
            ParentId = 9
        })
        $ps.Invoke()
        $ps.Streams.Progress
    }
    finally {
        $runspacePool.Dispose()
    }
    """

    with BackgroundPipeline(
        client_opened_pwsh,
        cmd,
        Name=server_pwsh.pipe_name,
    ) as ps:

        runspace = server_pwsh.runspace
        open_runspace(server_pwsh)

        command = server_pwsh.next_payload()
        assert command.action == "Command"
        server_pwsh.command_ack(command.ps_guid)

        s_ps = psrpcore.ServerPipeline(runspace, command.ps_guid)
        create_pipe = server_pwsh.next_event()
        assert isinstance(create_pipe, psrpcore.CreatePipelineEvent)
        assert isinstance(create_pipe.pipeline, psrpcore.PowerShell)
        assert create_pipe.pipeline_id == s_ps.pipeline_id
        assert len(create_pipe.pipeline.commands) == 1
        assert create_pipe.pipeline.no_input is True

        cmd1 = create_pipe.pipeline.commands[0]
        assert cmd1.command_text == "Write-Progress"
        assert cmd1.is_script is False
        assert cmd1.end_of_statement is True
        assert cmd1.parameters == [
            ("Activity", "act"),
            ("Status", "status"),
            ("Id", 10),
            ("PercentComplete", 34),
            ("SecondsRemaining", 102),
            ("CurrentOperation", "currentOp"),
            ("ParentId", 9),
        ]

        s_ps.start()
        s_ps.write_progress(
            activity="act",
            activity_id=10,
            status_description="status",
            current_operation="currentOp",
            parent_activity_id=9,
            percent_complete=34,
            seconds_remaining=102,
        )
        s_ps.complete()
        server_pwsh.data()

        close = server_pwsh.next_payload()
        server_pwsh.close_ack(close.ps_guid)

        progress_record = ps.events.get()
        assert isinstance(progress_record, psrpcore.PipelineOutputEvent)
        assert isinstance(progress_record.data, psrpcore.types.ProgressRecord)
        assert progress_record.data.Activity == "act"
        assert progress_record.data.ActivityId == 10
        assert progress_record.data.CurrentOperation == "currentOp"
        assert progress_record.data.ParentActivityId == 9
        assert progress_record.data.PercentComplete == 34
        assert progress_record.data.RecordType == psrpcore.types.ProgressRecordType.Processing
        assert progress_record.data.SecondsRemaining == 102
        assert progress_record.data.StatusDescription == "status"

        server_pwsh.next_payload()
        server_pwsh.close_ack(None)


def test_pipe_input_data(server_pwsh: ServerTransport, client_opened_pwsh: ClientTransport):
    cmd = """[CmdletBinding()]
    param ([String]$Name)

    $connInfo = [System.Management.Automation.Runspaces.NamedPipeConnectionInfo]::new($Name)
    $runspacePool = [RunspaceFactory]::CreateRunspacePool(1, 1, $connInfo)
    $runspacePool.Open()
    try {
        $ps = [PowerShell]::Create()
        $ps.RunspacePool = $runspacePool
        [void]$ps.AddScript('$input')
        $ps.Invoke(@(1, '2'))
    }
    finally {
        $runspacePool.Dispose()
    }
    """

    with BackgroundPipeline(
        client_opened_pwsh,
        cmd,
        Name=server_pwsh.pipe_name,
    ):

        runspace = server_pwsh.runspace
        open_runspace(server_pwsh)

        command = server_pwsh.next_payload()
        assert command.action == "Command"
        server_pwsh.command_ack(command.ps_guid)

        s_ps = psrpcore.ServerPipeline(runspace, command.ps_guid)
        create_pipe = server_pwsh.next_event()
        assert isinstance(create_pipe, psrpcore.CreatePipelineEvent)
        assert isinstance(create_pipe.pipeline, psrpcore.PowerShell)
        assert create_pipe.pipeline_id == s_ps.pipeline_id
        assert len(create_pipe.pipeline.commands) == 1
        assert create_pipe.pipeline.no_input is False

        cmd1 = create_pipe.pipeline.commands[0]
        assert cmd1.command_text == "$input"
        assert cmd1.is_script is True
        assert cmd1.end_of_statement is True
        assert cmd1.parameters == []

        server_pwsh.data_ack(s_ps.pipeline_id)
        s_ps.start()

        in_data = server_pwsh.next_event()
        assert isinstance(in_data, psrpcore.PipelineInputEvent)
        assert in_data.data == 1

        in_data = server_pwsh.next_event()
        assert isinstance(in_data, psrpcore.PipelineInputEvent)
        assert in_data.data == "2"

        in_end = server_pwsh.next_event()
        assert isinstance(in_end, psrpcore.EndOfPipelineInputEvent)

        server_pwsh.data_ack(s_ps.pipeline_id)
        s_ps.complete()
        server_pwsh.data()

        close = server_pwsh.next_payload()
        assert close.action == "Close"
        server_pwsh.close_ack(close.ps_guid)

        server_pwsh.next_payload()
        server_pwsh.close_ack(None)


@pytest.mark.skipif(os.name == "nt", reason="Very rare issue when the client fails to connect - no idea why.")
def test_stop_pipe(server_pwsh: ServerTransport, client_opened_pwsh: ClientTransport):
    cmd = """[CmdletBinding()]
    param ([String]$Name)

    $connInfo = [System.Management.Automation.Runspaces.NamedPipeConnectionInfo]::new($Name)
    $runspacePool = [RunspaceFactory]::CreateRunspacePool(1, 1, $connInfo)
    $runspacePool.Open()
    try {
        $ps = [PowerShell]::Create()
        $ps.RunspacePool = $runspacePool

        $out = [System.Management.Automation.PSDataCollection[PSObject]]::new()
        $in = [System.Management.Automation.PSDataCollection[PSObject]]::new(1)
        $in.Add(1)
        $in.Complete()

        Register-ObjectEvent -InputObject $out -EventName DataAdded -SourceIdentifier 'PSRPCore.Wait'

        [void]$ps.AddScript("echo 'out'; sleep 60")
        $task = $ps.BeginInvoke($in, $out)

        # Wait until the first object has been received back before sending the stop signal
        $null = Wait-Event -SourceIdentifier 'PSRPCore.Wait'
        $ps.Stop()

        $ps.EndInvoke($task)
    }
    finally {
        $runspacePool.Dispose()
    }
    """

    with BackgroundPipeline(
        client_opened_pwsh,
        cmd,
        Name=server_pwsh.pipe_name,
    ) as ps:

        runspace = server_pwsh.runspace
        open_runspace(server_pwsh)

        command = server_pwsh.next_payload()
        assert command.action == "Command"
        server_pwsh.command_ack(command.ps_guid)

        s_ps = psrpcore.ServerPipeline(runspace, command.ps_guid)

        create_pipe = server_pwsh.next_event()
        assert isinstance(create_pipe, psrpcore.CreatePipelineEvent)
        assert isinstance(create_pipe.pipeline, psrpcore.PowerShell)
        assert create_pipe.pipeline_id == s_ps.pipeline_id
        assert len(create_pipe.pipeline.commands) == 1
        server_pwsh.data_ack(s_ps.pipeline_id)

        cmd1 = create_pipe.pipeline.commands[0]
        assert cmd1.command_text == "echo 'out'; sleep 60"
        assert cmd1.is_script is True
        assert cmd1.end_of_statement is True
        assert cmd1.parameters == []

        s_ps.start()
        server_pwsh.data()

        # Get the input to know the pipeline has fully started on the client side
        input = server_pwsh.next_event()
        assert isinstance(input, psrpcore.PipelineInputEvent)
        assert input.pipeline_id == s_ps.pipeline_id
        assert input.data == 1

        eof = server_pwsh.next_event()
        assert isinstance(eof, psrpcore.EndOfPipelineInputEvent)
        server_pwsh.data_ack(s_ps.pipeline_id)

        # Send the output so the client knows it's fully started here
        s_ps.write_output("output")
        server_pwsh.data()

        # Now wait for the stop signal.
        signal = server_pwsh.next_payload()
        assert signal.action == "Signal"
        assert signal.ps_guid == s_ps.pipeline_id

        s_ps.stop()
        server_pwsh.data()

        server_pwsh.signal_ack(s_ps.pipeline_id)

        close = server_pwsh.next_payload()
        assert close.action == "Close"
        assert close.ps_guid == s_ps.pipeline_id
        server_pwsh.close_ack(close.ps_guid)

        close = server_pwsh.next_payload()
        assert close.action == "Close"
        assert close.ps_guid is None
        server_pwsh.close_ack(None)

        stop_record = ps.events.get()
        assert isinstance(stop_record, psrpcore.ErrorRecordEvent)
        assert (
            str(stop_record.record)
            == 'Exception calling "EndInvoke" with "1" argument(s): "The pipeline has been stopped."'
        )


def test_set_rp_count(server_pwsh: ServerTransport, client_opened_pwsh: ClientTransport):
    cmd = """[CmdletBinding()]
    param ([String]$Name)

    $connInfo = [System.Management.Automation.Runspaces.NamedPipeConnectionInfo]::new($Name)
    $runspacePool = [RunspaceFactory]::CreateRunspacePool(2, 5, $connInfo)
    $runspacePool.Open()
    try {
        $runspacePool.SetMaxRunspaces(1)  # $false
        $runspacePool.SetMaxRunspaces(5)  # $false

        $runspacePool.SetMaxRunspaces(4)  # $false - resp from server
        $runspacePool.SetMaxRunspaces(4)  # $true

        $runspacePool.SetMinRunspaces(6)  # $false - weird it sends it as 4 to the server
        $runspacePool.SetMinRunspaces(2)  # $false

        $runspacePool.SetMinRunspaces(4)  # $false - resp from server
        $runspacePool.SetMinRunspaces(4)  # $true

        $runspacePool.GetAvailableRunspaces()  # 3
    }
    finally {
        $runspacePool.Dispose()
    }
    """

    with BackgroundPipeline(
        client_opened_pwsh,
        cmd,
        Name=server_pwsh.pipe_name,
    ) as ps:
        open_runspace(server_pwsh)

        assert server_pwsh.runspace.min_runspaces == 2
        assert server_pwsh.runspace.max_runspaces == 5

        res = ps.events.get()
        assert isinstance(res, psrpcore.PipelineOutputEvent)
        assert res.data is False

        res = ps.events.get()
        assert isinstance(res, psrpcore.PipelineOutputEvent)
        assert res.data is False

        set_max = server_pwsh.next_event()
        assert isinstance(set_max, psrpcore.SetMaxRunspacesEvent)
        assert set_max.count == 4
        server_pwsh.data_ack()

        server_pwsh.runspace.runspace_availability_response(set_max.ci, False)
        server_pwsh.data()
        res = ps.events.get()
        assert isinstance(res, psrpcore.PipelineOutputEvent)
        assert res.data is False
        assert server_pwsh.runspace.max_runspaces == 5

        set_max = server_pwsh.next_event()
        assert isinstance(set_max, psrpcore.SetMaxRunspacesEvent)
        assert set_max.count == 4
        server_pwsh.data_ack()

        server_pwsh.runspace.runspace_availability_response(set_max.ci, True)
        server_pwsh.data()
        res = ps.events.get()
        assert isinstance(res, psrpcore.PipelineOutputEvent)
        assert res.data is True
        assert server_pwsh.runspace.max_runspaces == 4

        res = ps.events.get()
        assert isinstance(res, psrpcore.PipelineOutputEvent)
        assert res.data is False

        res = ps.events.get()
        assert isinstance(res, psrpcore.PipelineOutputEvent)
        assert res.data is False

        set_min = server_pwsh.next_event()
        assert isinstance(set_min, psrpcore.SetMinRunspacesEvent)
        assert set_min.count == 4
        server_pwsh.data_ack()

        server_pwsh.runspace.runspace_availability_response(set_min.ci, False)
        server_pwsh.data()
        res = ps.events.get()
        assert isinstance(res, psrpcore.PipelineOutputEvent)
        assert res.data is False
        assert server_pwsh.runspace.min_runspaces == 2

        set_min = server_pwsh.next_event()
        assert isinstance(set_min, psrpcore.SetMinRunspacesEvent)
        assert set_min.count == 4
        server_pwsh.data_ack()

        server_pwsh.runspace.runspace_availability_response(set_min.ci, True)
        server_pwsh.data()
        res = ps.events.get()
        assert isinstance(res, psrpcore.PipelineOutputEvent)
        assert res.data is True
        assert server_pwsh.runspace.min_runspaces == 4

        get_avail = server_pwsh.next_event()
        assert isinstance(get_avail, psrpcore.GetAvailableRunspacesEvent)
        server_pwsh.data_ack()

        server_pwsh.runspace.runspace_availability_response(get_avail.ci, 3)
        server_pwsh.data()

        res = ps.events.get()
        assert isinstance(res, psrpcore.PipelineOutputEvent)
        assert res.data == 3

        close = server_pwsh.next_payload()
        assert close.action == "Close"
        server_pwsh.close_ack()


@pytest.mark.skipif(os.name == "nt", reason="Very rare issue when the client fails to connect - no idea why.")
def test_reset_rp(server_pwsh: ServerTransport, client_opened_pwsh: ClientTransport):
    cmd = """[CmdletBinding()]
    param ([String]$Name)

    $ErrorActionPreference = 'Stop'

    $connInfo = [System.Management.Automation.Runspaces.NamedPipeConnectionInfo]::new($Name)
    $runspace = [RunspaceFactory]::CreateRunspace($connInfo)
    $runspace.Open()
    try {
        $runspace.ResetRunspaceState()
        $runspace.ResetRunspaceState()
    }
    finally {
        $runspace.Dispose()
    }
    """

    with BackgroundPipeline(
        client_opened_pwsh,
        cmd,
        Name=server_pwsh.pipe_name,
    ) as ps:
        open_runspace(server_pwsh)

        reset = server_pwsh.next_event()
        assert isinstance(reset, psrpcore.ResetRunspaceStateEvent)
        server_pwsh.runspace.runspace_availability_response(reset.ci, True)
        server_pwsh.data()
        server_pwsh.data_ack()

        reset = server_pwsh.next_event()
        assert isinstance(reset, psrpcore.ResetRunspaceStateEvent)
        server_pwsh.runspace.runspace_availability_response(reset.ci, False)
        server_pwsh.data()
        server_pwsh.data_ack()

        close = server_pwsh.next_payload()
        assert close.action == "Close"
        server_pwsh.close_ack()

        state = ps.events.get()

        assert isinstance(state, psrpcore.PipelineStateEvent)
        assert isinstance(state.reason, psrpcore.types.ErrorRecord)
        assert state.state == psrpcore.types.RunspacePoolState.Broken
        assert '"ResetRunspaceState" is not valid' in str(state.reason)


def test_pipe_host_call(server_pwsh: ServerTransport, client_opened_pwsh: ClientTransport):
    cmd = """[CmdletBinding()]
    param ([String]$Name)

    $connInfo = [System.Management.Automation.Runspaces.NamedPipeConnectionInfo]::new($Name)
    $runspacePool = [RunspaceFactory]::CreateRunspacePool(1, 1, $connInfo, $Host)
    $runspacePool.Open()
    try {
        $ps = [PowerShell]::Create()
        $ps.RunspacePool = $runspacePool
        [void]$ps.AddScript(@'
$host.UI.WriteLine("line")
$host.UI.ReadLineAsSecureString()
'@)

        $ps.Invoke()
    }
    finally {
        $runspacePool.Dispose()
    }
    """

    with BackgroundPipeline(
        client_opened_pwsh,
        cmd,
        Name=server_pwsh.pipe_name,
    ) as ps:
        open_runspace(server_pwsh)
        runspace = server_pwsh.runspace

        command = server_pwsh.next_payload()
        assert command.action == "Command"
        server_pwsh.command_ack(command.ps_guid)

        s_ps = psrpcore.ServerPipeline(runspace, command.ps_guid)
        create_pipe = server_pwsh.next_event()
        assert isinstance(create_pipe, psrpcore.CreatePipelineEvent)
        assert isinstance(create_pipe.pipeline, psrpcore.PowerShell)
        assert create_pipe.pipeline_id == s_ps.pipeline_id
        server_pwsh.data_ack(s_ps.pipeline_id)
        s_ps.start()
        server_pwsh.data()

        s_host = psrpcore.ServerHostRequestor(s_ps)
        ci = s_host.write_line("line")
        assert ci is None
        server_pwsh.data()

        ci = s_host.read_line_as_secure_string()
        assert ci == 1
        server_pwsh.data()

        call1 = ps.events.get()
        assert isinstance(call1, psrpcore.PipelineHostCallEvent)
        assert call1.ci == -100
        assert call1.method_identifier == psrpcore.types.HostMethodIdentifier.WriteLine2
        assert call1.method_parameters == ["line"]

        record = ps.events.get()
        assert isinstance(record, psrpcore.WarningRecordEvent)
        assert "is asking to read a line securely" in record.record.Message

        call2 = ps.events.get()
        assert isinstance(call2, psrpcore.PipelineHostCallEvent)
        assert call2.ci == -100
        assert call2.method_identifier == psrpcore.types.HostMethodIdentifier.WriteWarningLine
        assert call2.method_parameters == [record.record.Message]

        call3 = ps.events.get()
        assert isinstance(call3, psrpcore.PipelineHostCallEvent)
        assert call3.ci == 1
        assert call3.method_identifier == psrpcore.types.HostMethodIdentifier.ReadLineAsSecureString
        assert call3.method_parameters == []

        client_opened_pwsh.runspace.exchange_key()
        client_opened_pwsh.data()
        enc_key = ps.events.get()
        assert isinstance(enc_key, psrpcore.EncryptedSessionKeyEvent)

        c_host = psrpcore.ClientHostResponder(ps.pipeline)
        c_host.read_line_as_secure_string(call3.ci, psrpcore.types.PSSecureString("secret"))
        client_opened_pwsh.data()

        pub_key = server_pwsh.next_event()
        assert isinstance(pub_key, psrpcore.PublicKeyEvent)
        server_pwsh.data()

        resp = server_pwsh.next_event()
        assert isinstance(resp, psrpcore.PipelineHostResponseEvent)
        assert resp.ci == 1
        assert resp.method_identifier == psrpcore.types.HostMethodIdentifier.ReadLineAsSecureString
        assert isinstance(resp.result, psrpcore.types.PSSecureString)
        assert resp.result.decrypt() == "secret"

        s_ps.complete()
        server_pwsh.data()

        close = server_pwsh.next_payload()
        assert close.action == "Close"
        assert close.ps_guid == s_ps.pipeline_id
        server_pwsh.close_ack(close.ps_guid)
        s_ps.close()
        server_pwsh.data()

        close = server_pwsh.next_payload()
        assert close.action == "Close"
        assert close.ps_guid is None
        server_pwsh.close_ack()


def test_rp_host_call(server_pwsh: ServerTransport, client_opened_pwsh: ClientTransport):
    cmd = """[CmdletBinding()]
    param ([String]$Name)

    $connInfo = [System.Management.Automation.Runspaces.NamedPipeConnectionInfo]::new($Name)
    $runspacePool = [RunspaceFactory]::CreateRunspacePool(1, 1, $connInfo, $Host)
    $runspacePool.Open()
    try {
        $ps = [PowerShell]::Create()
        $ps.RunspacePool = $runspacePool
        [void]$ps.AddScript(@'
$rs = [Runspace]::DefaultRunspace
$rsHost = $rs.GetType().GetProperty("Host", 60).GetValue($rs)
$rsHost.UI.WriteWarningLine("test")
'@)

        $ps.Invoke()
    }
    finally {
        $runspacePool.Dispose()
    }
    """

    with BackgroundPipeline(
        client_opened_pwsh,
        cmd,
        Name=server_pwsh.pipe_name,
    ) as ps:
        open_runspace(server_pwsh)
        runspace = server_pwsh.runspace

        command = server_pwsh.next_payload()
        assert command.action == "Command"
        server_pwsh.command_ack(command.ps_guid)

        s_ps = psrpcore.ServerPipeline(runspace, command.ps_guid)
        create_pipe = server_pwsh.next_event()
        assert isinstance(create_pipe, psrpcore.CreatePipelineEvent)
        assert isinstance(create_pipe.pipeline, psrpcore.PowerShell)
        assert create_pipe.pipeline_id == s_ps.pipeline_id
        server_pwsh.data_ack(s_ps.pipeline_id)
        s_ps.start()
        server_pwsh.data()

        s_host = psrpcore.ServerHostRequestor(runspace)
        ci = s_host.write_warning_line("test")
        assert ci is None
        server_pwsh.data()

        # By the time it reaches our client (one controlling the pssession) it has turned into a warning record and
        # pipeline host call.
        record = ps.events.get()
        assert isinstance(record, psrpcore.WarningRecordEvent)
        assert record.record.Message == "test"
        call1 = ps.events.get()
        assert isinstance(call1, psrpcore.PipelineHostCallEvent)
        assert call1.ci == -100
        assert call1.method_identifier == psrpcore.types.HostMethodIdentifier.WriteWarningLine
        assert call1.method_parameters == ["test"]

        s_ps.complete()
        server_pwsh.data()

        close = server_pwsh.next_payload()
        assert close.action == "Close"
        assert close.ps_guid == s_ps.pipeline_id
        server_pwsh.close_ack(close.ps_guid)
        s_ps.close()
        server_pwsh.data()

        close = server_pwsh.next_payload()
        assert close.action == "Close"
        assert close.ps_guid is None
        server_pwsh.close_ack()


def test_cmd_meta(server_pwsh: ServerTransport, client_opened_pwsh: ClientTransport):
    cmd = """[CmdletBinding()]
    param ([String]$Name)

    [System.Management.Automation.Platform]::SelectProductNameForDirectory('USER_MODULES')

    $connInfo = [System.Management.Automation.Runspaces.NamedPipeConnectionInfo]::new($Name)
    $runspace = [RunspaceFactory]::CreateRunspace($connInfo)
    $runspace.Open()
    try {
        $cstr = [Management.Automation.Runspaces.PSSession].GetConstructor(
            'NonPublic, Instance', $null, [type[]]$runspace.GetType(), $null)
        $session = $cstr.Invoke(@($runspace))

        $exportParams = @{
            Session = $session
            OutputModule = 'psrpcore-testing'
            CommandName = 'Get-*Item'
            CommandType = 'Cmdlet'
            ArgumentList = 'env:'
            Force = $true
        }
        Export-PSSession @exportParams
    }
    finally {
        $runspace.Dispose()
    }
    """

    module_path = None
    try:
        with BackgroundPipeline(
            client_opened_pwsh,
            cmd,
            Name=server_pwsh.pipe_name,
        ) as ps:
            module_path = ps.events.get().data

            open_runspace(server_pwsh)
            runspace = server_pwsh.runspace

            command = server_pwsh.next_payload()
            assert command.action == "Command"
            server_pwsh.command_ack(command.ps_guid)

            s_ps = psrpcore.ServerPipeline(runspace, command.ps_guid)
            cmd_meta = server_pwsh.next_event()
            assert isinstance(cmd_meta, psrpcore.GetCommandMetadataEvent)
            assert isinstance(cmd_meta.pipeline, psrpcore.GetMetadata)
            assert cmd_meta.pipeline_id == s_ps.pipeline_id
            assert cmd_meta.pipeline.name == ["Get-*Item"]
            assert cmd_meta.pipeline.command_type == psrpcore.types.CommandTypes.Cmdlet
            assert cmd_meta.pipeline.namespace == []
            assert cmd_meta.pipeline.arguments == ["env:"]

            server_pwsh.data_ack(s_ps.pipeline_id)
            s_ps.start()
            server_pwsh.data()

            s_ps.write_output(
                psrpcore.types.PSCustomObject(
                    PSTypeName="Selected.Microsoft.PowerShell.Commands.GenericMeasureInfo",
                    Count=2,
                )
            )
            server_pwsh.data()

            s_ps.write_output(
                psrpcore.types.PSCustomObject(
                    PSTypeName="Selected.System.Management.Automation.CmdletInfo",
                    Name="Get-ChildItem",
                    Namespace="Microsoft.PowerShell.Management",
                    HelpUri="https://go.microsoft.com/fwlink/?LinkID=2096492",
                    CommandType=psrpcore.types.CommandTypes.Cmdlet,
                    ResolvedCommandName=None,
                    OutputType=["System.IO.FileInfo", "System.IO.DirectoryInfo"],
                    Parameters={},
                )
            )
            server_pwsh.data()

            s_ps.write_output(
                psrpcore.types.PSCustomObject(
                    PSTypeName="Selected.System.Management.Automation.CmdletInfo",
                    Name="Get-Item",
                    Namespace="Microsoft.PowerShell.Management",
                    HelpUri="https://go.microsoft.com/fwlink/?LinkID=2096492",
                    CommandType=psrpcore.types.CommandTypes.Cmdlet,
                    ResolvedCommandName=None,
                    OutputType=["System.IO.FileInfo", "System.Boolean", "System.String"],
                    Parameters={},
                )
            )
            server_pwsh.data()

            s_ps.complete()
            server_pwsh.data()

            close = server_pwsh.next_payload()
            assert close.action == "Close"
            assert close.ps_guid == s_ps.pipeline_id
            server_pwsh.close_ack(close.ps_guid)
            s_ps.close()
            server_pwsh.data()

            close = server_pwsh.next_payload()
            assert close.action == "Close"
            assert close.ps_guid is None
            server_pwsh.close_ack()

    finally:
        if not module_path:
            return

        exported_path = os.path.join(module_path, "psrpcore-testing")
        if os.path.exists(exported_path):
            shutil.rmtree(exported_path)


def test_user_event(server_pwsh: ServerTransport, client_opened_pwsh: ClientTransport):
    cmd = """[CmdletBinding()]
    param ([String]$Name)

    $connInfo = [System.Management.Automation.Runspaces.NamedPipeConnectionInfo]::new($Name)
    $runspace = [RunspaceFactory]::CreateRunspace($Host, $connInfo)
    $runspace.Open()
    try {
        $eventParams = @{
            InputObject = $runspace.Events.ReceivedEvents
            EventName = "PSEventReceived"
            SourceIdentifier = "PSRPCore.UserEvent"
        }
        $null = Register-ObjectEvent @eventParams

        $ps = [PowerShell]::Create()
        $ps.Runspace = $runspace
        [void]$ps.AddScript(@'
$null = $Host.Runspace.Events.SubscribeEvent(
    $null,
    "PSRPCoreEvent",
    "PSRPCoreEvent",
    $null,
    $null,
    $true,
    $true)
$null = $Host.Runspace.Events.GenerateEvent(
    "PSRPCoreEvent",
    "sender",
    @("my", "args"),
    "extra data")
'@)
        $ps.Invoke()

        Wait-Event -SourceIdentifier $eventParams.SourceIdentifier | Select *
    }
    finally {
        $runspace.Dispose()
    }
    """

    with BackgroundPipeline(
        client_opened_pwsh,
        cmd,
        Name=server_pwsh.pipe_name,
    ) as ps:
        open_runspace(server_pwsh)
        runspace = server_pwsh.runspace

        command = server_pwsh.next_payload()
        assert command.action == "Command"
        server_pwsh.command_ack(command.ps_guid)

        s_ps = psrpcore.ServerPipeline(runspace, command.ps_guid)
        create_pipe = server_pwsh.next_event()
        assert isinstance(create_pipe, psrpcore.CreatePipelineEvent)
        assert isinstance(create_pipe.pipeline, psrpcore.PowerShell)
        assert create_pipe.pipeline_id == s_ps.pipeline_id

        server_pwsh.data_ack(s_ps.pipeline_id)
        s_ps.start()
        server_pwsh.data()

        runspace.send_event(
            event_identifier=1,
            source_identifier="PSRPCoreEvent",
            sender="sender",
            source_args=["my 1", "args 2"],
            message_data="extra data",
        )
        server_pwsh.data()

        s_ps.complete()
        server_pwsh.data()

        close = server_pwsh.next_payload()
        assert close.action == "Close"
        assert close.ps_guid == s_ps.pipeline_id
        server_pwsh.close_ack(close.ps_guid)

        close = server_pwsh.next_payload()
        assert close.action == "Close"
        assert close.ps_guid is None
        server_pwsh.close_ack()

        event = ps.events.get()
        assert isinstance(event, psrpcore.PipelineOutputEvent)
        assert event.data.EventIdentifier == 1
        assert event.data.SourceIdentifier == "PSRPCore.UserEvent"

        source_event = event.data.SourceEventArgs
        assert source_event.EventIdentifier == 1
        assert source_event.MessageData == "extra data"
        assert source_event.Sender == "sender"
        assert source_event.SourceArgs == ["my 1", "args 2"]
        assert source_event.SourceEventArgs is None
        assert source_event.SourceIdentifier == "PSRPCoreEvent"


def test_set_buffer_call(server_pwsh: ServerTransport, client_opened_pwsh: ClientTransport):
    cmd = """[CmdletBinding()]
    param ([String]$Name)

    $connInfo = [System.Management.Automation.Runspaces.NamedPipeConnectionInfo]::new($Name)
    $runspace = [RunspaceFactory]::CreateRunspace($Host, $connInfo)
    $runspace.Open()
    try {
        $ps = [PowerShell]::Create()
        $ps.Runspace = $runspace
        [void]$ps.AddScript(@'
$coordinates = [System.Management.Automation.Host.Coordinates]::new(0, 1)
$cell = [System.Management.Automation.Host.BufferCell]::new('a', 'White', 'Gray', 'Complete')
$cells = $Host.UI.RawUI.NewBufferCellArray(3, 4, $cell)
$host.UI.RawUI.SetBufferContents($coordinates, $cells)
'@)
        $ps.Invoke()
    }
    finally {
        $runspace.Dispose()
    }
    """

    with BackgroundPipeline(
        client_opened_pwsh,
        cmd,
        Name=server_pwsh.pipe_name,
    ) as ps:
        open_runspace(server_pwsh)
        runspace = server_pwsh.runspace

        command = server_pwsh.next_payload()
        assert command.action == "Command"
        server_pwsh.command_ack(command.ps_guid)

        s_ps = psrpcore.ServerPipeline(runspace, command.ps_guid)
        create_pipe = server_pwsh.next_event()
        assert isinstance(create_pipe, psrpcore.CreatePipelineEvent)
        assert isinstance(create_pipe.pipeline, psrpcore.PowerShell)
        assert create_pipe.pipeline_id == s_ps.pipeline_id
        server_pwsh.data_ack(s_ps.pipeline_id)

        s_ps.start()
        server_pwsh.data()

        s_host = psrpcore.ServerHostRequestor(s_ps)
        cell = psrpcore.types.BufferCell(
            "a",
            psrpcore.types.ConsoleColor.White,
            psrpcore.types.ConsoleColor.Black,
            psrpcore.types.BufferCellType.Complete,
        )
        contents = [
            [cell, cell, cell],
            [cell, cell, cell],
            [cell, cell, cell],
            [cell, cell, cell],
        ]
        ci = s_host.set_buffer_contents(0, 1, contents)
        # ci = s_host.set_buffer_cells(0, 1, 4, 5, "a")
        assert ci is None
        server_pwsh.data()

        s_ps.complete()
        server_pwsh.data()

        close = server_pwsh.next_payload()
        assert close.action == "Close"
        assert close.ps_guid == s_ps.pipeline_id
        server_pwsh.close_ack(close.ps_guid)

        close = server_pwsh.next_payload()
        assert close.action == "Close"
        assert close.ps_guid is None
        server_pwsh.close_ack()

        host_call = ps.events.get()
        assert isinstance(host_call, psrpcore.PipelineHostCallEvent)
        assert host_call.ci == -100
        assert host_call.method_identifier == psrpcore.types.HostMethodIdentifier.SetBufferContents2
        assert len(host_call.method_parameters) == 2

        coordinates = host_call.method_parameters[0]
        assert isinstance(coordinates, psrpcore.types.Coordinates)
        assert coordinates.X == 0
        assert coordinates.Y == 1

        contents = host_call.method_parameters[1]
        assert isinstance(contents, list)
        assert len(contents) == 4
        for row in contents:
            assert isinstance(row, list)
            assert len(row) == 3
            for cell in row:
                assert isinstance(cell, psrpcore.types.BufferCell)
                assert cell.Character == psrpcore.types.PSChar("a")
                assert cell.ForegroundColor == psrpcore.types.ConsoleColor.White
                assert isinstance(cell.ForegroundColor, psrpcore.types.ConsoleColor)
                assert cell.BackgroundColor == psrpcore.types.ConsoleColor.Black
                assert isinstance(cell.BackgroundColor, psrpcore.types.ConsoleColor)
                assert cell.BufferCellType == psrpcore.types.BufferCellType.Complete


def test_pipeline_extra_cmds(server_pwsh: ServerTransport, client_opened_pwsh: ClientTransport):
    cmd = """[CmdletBinding()]
    param ([String]$Name)

    $connInfo = [System.Management.Automation.Runspaces.NamedPipeConnectionInfo]::new($Name)
    $runspace = [RunspaceFactory]::CreateRunspace($Host, $connInfo)
    $runspace.Open()
    try {
        $ps = [PowerShell]::Create()
        $ps.Runspace = $runspace

        [void]$ps.AddCommand("Set-Variable").AddParameters([Ordered]@{
            Name = "string"
            Value = "foo"
        }).AddStatement()

        [void]$ps.AddCommand("Get-Variable").AddParameter("Name", "string")
        [void]$ps.AddCommand("Select-Object").AddParameter("Property", @("Name", "Value"))
        [void]$ps.AddStatement()

        [void]$ps.AddCommand("Get-Variable").AddArgument("string").AddParameter("ValueOnly", $true)
        [void]$ps.AddCommand("Select-Object")
        [void]$ps.AddStatement()

        [void]$ps.AddScript('[PSCustomObject]@{ Value = $string }')
        [void]$ps.AddScript('process { $_ | Select-Object -Property @{N="Test"; E={ $_.Value }} }')
        [void]$ps.AddStatement()

        $ps.Invoke()
    }
    finally {
        $runspace.Dispose()
    }
    """

    with BackgroundPipeline(
        client_opened_pwsh,
        cmd,
        Name=server_pwsh.pipe_name,
    ) as ps:
        open_runspace(server_pwsh)
        runspace = server_pwsh.runspace

        command = server_pwsh.next_payload()
        assert command.action == "Command"
        server_pwsh.command_ack(command.ps_guid)

        s_ps = psrpcore.ServerPipeline(runspace, command.ps_guid)
        create_pipe = server_pwsh.next_event()
        assert isinstance(create_pipe, psrpcore.CreatePipelineEvent)
        assert isinstance(create_pipe.pipeline, psrpcore.PowerShell)
        assert create_pipe.pipeline_id == s_ps.pipeline_id
        assert len(s_ps.metadata.commands) == 7
        assert s_ps.metadata.commands[0].command_text == "Set-Variable"
        assert s_ps.metadata.commands[0].parameters == [("Name", "string"), ("Value", "foo")]
        assert s_ps.metadata.commands[0].end_of_statement
        assert s_ps.metadata.commands[0].is_script is False

        assert s_ps.metadata.commands[1].command_text == "Get-Variable"
        assert s_ps.metadata.commands[1].parameters == [("Name", "string")]
        assert s_ps.metadata.commands[1].end_of_statement is False
        assert s_ps.metadata.commands[1].is_script is False

        assert s_ps.metadata.commands[2].command_text == "Select-Object"
        assert s_ps.metadata.commands[2].parameters == [("Property", ["Name", "Value"])]
        assert s_ps.metadata.commands[2].end_of_statement
        assert s_ps.metadata.commands[2].is_script is False

        assert s_ps.metadata.commands[3].command_text == "Get-Variable"
        assert s_ps.metadata.commands[3].parameters == [(None, "string"), ("ValueOnly", True)]
        assert s_ps.metadata.commands[3].end_of_statement is False
        assert s_ps.metadata.commands[3].is_script is False

        assert s_ps.metadata.commands[4].command_text == "Select-Object"
        assert s_ps.metadata.commands[4].parameters == []
        assert s_ps.metadata.commands[4].end_of_statement
        assert s_ps.metadata.commands[4].is_script is False

        assert s_ps.metadata.commands[5].command_text == "[PSCustomObject]@{ Value = $string }"
        assert s_ps.metadata.commands[5].parameters == []
        assert s_ps.metadata.commands[5].end_of_statement is False
        assert s_ps.metadata.commands[5].is_script

        assert (
            s_ps.metadata.commands[6].command_text
            == 'process { $_ | Select-Object -Property @{N="Test"; E={ $_.Value }} }'
        )
        assert s_ps.metadata.commands[6].parameters == []
        assert s_ps.metadata.commands[6].end_of_statement is True
        assert s_ps.metadata.commands[6].is_script

        server_pwsh.data_ack(s_ps.pipeline_id)
        s_ps.start()
        server_pwsh.data()

        s_ps.complete()
        server_pwsh.data()

        close = server_pwsh.next_payload()
        assert close.action == "Close"
        assert close.ps_guid == s_ps.pipeline_id
        server_pwsh.close_ack(close.ps_guid)

        close = server_pwsh.next_payload()
        assert close.action == "Close"
        assert close.ps_guid is None
        server_pwsh.close_ack()



# -*- coding: utf-8 -*-
# Copyright: (c) 2021, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import datetime
import xml.etree.ElementTree as ElementTree

import psrpcore.types._psrp as psrp
from psrpcore.types import (
    ApartmentState,
    DebugRecord,
    ErrorCategory,
    ErrorCategoryInfo,
    HostInfo,
    NETException,
    PSBool,
    PSDateTime,
    PSGuid,
    PSInt,
    PSInt64,
    PSObject,
    PSString,
    PSThreadOptions,
    PSVersion,
)

from ..conftest import assert_xml_diff, deserialize, serialize


def test_session_capability():
    ps_value = psrp.SessionCapability("1.2", "1.2.3", PSVersion("4.5.6.7"))
    assert ps_value.PSVersion == PSVersion("1.2")
    assert ps_value.protocolversion == PSVersion("1.2.3")
    assert ps_value.SerializationVersion == PSVersion("4.5.6.7")
    assert ps_value.TimeZone is None

    element = serialize(ps_value)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert (
        actual == '<Obj RefId="0">'
        "<MS>"
        '<Version N="PSVersion">1.2</Version>'
        '<Version N="protocolversion">1.2.3</Version>'
        '<Version N="SerializationVersion">4.5.6.7</Version>'
        "</MS>"
        "</Obj>"
    )

    ps_value = deserialize(element)
    assert isinstance(ps_value, PSObject)
    assert not isinstance(ps_value, psrp.SessionCapability)
    assert ps_value.PSVersion == PSVersion("1.2")
    assert ps_value.protocolversion == PSVersion("1.2.3")
    assert ps_value.SerializationVersion == PSVersion("4.5.6.7")

    # Because we couldn't rehydrate the object there is no TimeZone property.
    assert len(ps_value.PSObject.extended_properties) == 3


def test_session_capability_with_timezone():
    ps_value = psrp.SessionCapability("1.2", "1.2.3", TimeZone=b"\x00\x01\x02\x03", SerializationVersion="4.5")
    assert ps_value.PSVersion == PSVersion("1.2")
    assert ps_value.protocolversion == PSVersion("1.2.3")
    assert ps_value.SerializationVersion == PSVersion("4.5")
    assert ps_value.TimeZone == b"\x00\x01\x02\x03"

    element = serialize(ps_value)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert (
        actual == '<Obj RefId="0">'
        "<MS>"
        '<Version N="PSVersion">1.2</Version>'
        '<Version N="protocolversion">1.2.3</Version>'
        '<Version N="SerializationVersion">4.5</Version>'
        '<BA N="TimeZone">AAECAw==</BA>'
        "</MS>"
        "</Obj>"
    )

    ps_value = deserialize(element)
    assert isinstance(ps_value, PSObject)
    assert not isinstance(ps_value, psrp.SessionCapability)
    assert ps_value.PSVersion == PSVersion("1.2")
    assert ps_value.protocolversion == PSVersion("1.2.3")
    assert ps_value.SerializationVersion == PSVersion("4.5")
    assert ps_value.TimeZone == b"\x00\x01\x02\x03"
    assert len(ps_value.PSObject.extended_properties) == 4


def test_init_runspace_pool():
    ps_value = psrp.InitRunspacePool(
        1, 2, PSThreadOptions.UseNewThread, ApartmentState.STA, HostInfo(), {"key": "value"}
    )
    assert isinstance(ps_value, psrp.InitRunspacePool)
    assert ps_value.MinRunspaces == 1
    assert ps_value.MaxRunspaces == 2
    assert ps_value.PSThreadOptions == PSThreadOptions.UseNewThread
    assert ps_value.ApartmentState == ApartmentState.STA
    assert isinstance(ps_value.HostInfo, HostInfo)
    assert ps_value.HostInfo.IsHostNull
    assert ps_value.HostInfo.IsHostUINull
    assert ps_value.HostInfo.IsHostRawUINull
    assert ps_value.HostInfo.UseRunspaceHost
    assert ps_value.HostInfo.HostDefaultData is None
    assert ps_value.ApplicationArguments == {"key": "value"}

    element = serialize(ps_value)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    expected = (
        '<Obj RefId="0">'
        "<MS>"
        '<I32 N="MinRunspaces">1</I32>'
        '<I32 N="MaxRunspaces">2</I32>'
        '<Obj RefId="1" N="PSThreadOptions">'
        "<I32>1</I32>"
        '<TN RefId="0">'
        "<T>System.Management.Automation.Runspaces.PSThreadOptions</T>"
        "<T>System.Enum</T>"
        "<T>System.ValueType</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<ToString>UseNewThread</ToString>"
        "</Obj"
        '><Obj RefId="2" N="ApartmentState">'
        "<I32>0</I32>"
        '<TN RefId="1">'
        "<T>System.Threading.ApartmentState</T>"
        "<T>System.Enum</T>"
        "<T>System.ValueType</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<ToString>STA</ToString>"
        "</Obj>"
        '<Obj RefId="3" N="HostInfo">'
        "<MS>"
        '<B N="_isHostNull">true</B>'
        '<B N="_isHostUINull">true</B>'
        '<B N="_isHostRawUINull">true</B>'
        '<B N="_useRunspaceHost">true</B>'
        "</MS>"
        "</Obj>"
        '<Obj RefId="4" N="ApplicationArguments">'
        '<TN RefId="2">'
        "<T>System.Management.Automation.PSPrimitiveDictionary</T>"
        "<T>System.Collections.Hashtable</T>"
        "<T>System.Object</T>"
        "</TN>"
        '<DCT><En><S N="Key">key</S><S N="Value">value</S></En></DCT>'
        "</Obj>"
        "</MS>"
        "</Obj>"
    )
    assert_xml_diff(actual, expected)

    ps_value = deserialize(element)
    assert isinstance(ps_value, PSObject)
    assert not isinstance(ps_value, psrp.InitRunspacePool)
    assert ps_value.MinRunspaces == 1
    assert ps_value.MaxRunspaces == 2
    assert ps_value.PSThreadOptions == PSThreadOptions.UseNewThread
    assert ps_value.ApartmentState == ApartmentState.STA
    assert ps_value.HostInfo._isHostUINull
    assert ps_value.HostInfo._isHostRawUINull
    assert ps_value.HostInfo._useRunspaceHost
    assert not hasattr(ps_value.HostInfo, "_hostDefaultData")
    assert ps_value.ApplicationArguments == {"key": "value"}


def test_public_key():
    ps_value = psrp.PublicKey("test")
    assert isinstance(ps_value, psrp.PublicKey)
    assert ps_value.PublicKey == "test"
    assert isinstance(ps_value.PublicKey, PSString)

    element = serialize(ps_value)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert actual == '<Obj RefId="0"><MS><S N="PublicKey">test</S></MS></Obj>'

    ps_value = deserialize(element)
    assert isinstance(ps_value, PSObject)
    assert not isinstance(ps_value, psrp.PublicKey)
    assert isinstance(ps_value.PublicKey, PSString)
    assert ps_value.PublicKey == "test"


def test_encrypted_session_key():
    ps_value = psrp.EncryptedSessionKey("test")
    assert isinstance(ps_value, psrp.EncryptedSessionKey)
    assert ps_value.EncryptedSessionKey == "test"
    assert isinstance(ps_value.EncryptedSessionKey, PSString)

    element = serialize(ps_value)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert actual == '<Obj RefId="0"><MS><S N="EncryptedSessionKey">test</S></MS></Obj>'

    ps_value = deserialize(element)
    assert isinstance(ps_value, PSObject)
    assert not isinstance(ps_value, psrp.EncryptedSessionKey)
    assert isinstance(ps_value.EncryptedSessionKey, PSString)
    assert ps_value.EncryptedSessionKey == "test"


def test_public_key_request():
    ps_value = psrp.PublicKeyRequest()
    assert isinstance(ps_value, psrp.PublicKeyRequest)
    assert ps_value == ""

    element = serialize(ps_value)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert actual == "<S />"

    ps_value = deserialize(element)
    assert isinstance(ps_value, PSString)
    assert not isinstance(ps_value, psrp.EncryptedSessionKey)
    assert ps_value == ""


def test_set_max_runspaces():
    ps_value = psrp.SetMaxRunspaces(10, -20)
    assert isinstance(ps_value, psrp.SetMaxRunspaces)
    assert ps_value.MaxRunspaces == 10
    assert ps_value.ci == -20

    element = serialize(ps_value)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert actual == '<Obj RefId="0">' "<MS>" '<I32 N="MaxRunspaces">10</I32>' '<I64 N="ci">-20</I64>' "</MS>" "</Obj>"

    ps_value = deserialize(element)
    assert isinstance(ps_value, PSObject)
    assert not isinstance(ps_value, psrp.SetMaxRunspaces)
    assert isinstance(ps_value.MaxRunspaces, PSInt)
    assert ps_value.MaxRunspaces == 10
    assert isinstance(ps_value.ci, PSInt64)
    assert ps_value.ci == -20


def test_set_min_runspaces():
    ps_value = psrp.SetMinRunspaces(10, -20)
    assert isinstance(ps_value, psrp.SetMinRunspaces)
    assert ps_value.MinRunspaces == 10
    assert ps_value.ci == -20

    element = serialize(ps_value)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert actual == '<Obj RefId="0">' "<MS>" '<I32 N="MinRunspaces">10</I32>' '<I64 N="ci">-20</I64>' "</MS>" "</Obj>"

    ps_value = deserialize(element)
    assert isinstance(ps_value, PSObject)
    assert not isinstance(ps_value, psrp.SetMinRunspaces)
    assert isinstance(ps_value.MinRunspaces, PSInt)
    assert ps_value.MinRunspaces == 10
    assert isinstance(ps_value.ci, PSInt64)
    assert ps_value.ci == -20


def tet_runspace_availability_bool():
    ps_value = psrp.RunspaceAvailability(True, 50)
    assert isinstance(ps_value, psrp.RunspaceAvailability)
    assert ps_value.SetMinMaxRunspacesResponse is True
    assert ps_value.ci == 50

    element = serialize(ps_value)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert (
        actual == '<Obj RefId="0">'
        "<MS>"
        '<B N="SetMinMaxRunspacesResponse">true</B>'
        '<I64 N="ci">50</I64>'
        "</MS>"
        "</Obj>"
    )

    ps_value = deserialize(element)
    assert isinstance(ps_value, PSObject)
    assert not isinstance(ps_value, psrp.RunspaceAvailability)
    assert isinstance(ps_value.SetMinMaxRunspacesResponse, PSBool)
    assert ps_value.SetMinMaxRunspacesResponse is True
    assert isinstance(ps_value.ci, PSInt64)
    assert ps_value.ci == 50


def tet_runspace_availability_long():
    ps_value = psrp.RunspaceAvailability(PSInt64(10), 50)
    assert isinstance(ps_value, psrp.RunspaceAvailability)
    assert ps_value.SetMinMaxRunspacesResponse == 10
    assert ps_value.ci == 50

    element = serialize(ps_value)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert (
        actual == '<Obj RefId="0">'
        "<MS>"
        '<I64 N="SetMinMaxRunspacesResponse">10</I64>'
        '<I64 N="ci">50</I64>'
        "</MS>"
        "</Obj>"
    )

    ps_value = deserialize(element)
    assert isinstance(ps_value, PSObject)
    assert not isinstance(ps_value, psrp.RunspaceAvailability)
    assert isinstance(ps_value.SetMinMaxRunspacesResponse, PSInt64)
    assert ps_value.SetMinMaxRunspacesResponse == 10
    assert isinstance(ps_value.ci, PSInt64)
    assert ps_value.ci == 50


def test_runspace_pool_state():
    ps_value = psrp.RunspacePoolStateMsg(1)
    assert ps_value.RunspaceState == 1
    assert isinstance(ps_value.RunspaceState, PSInt)
    assert ps_value.ExceptionAsErrorRecord is None

    element = serialize(ps_value)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert actual == '<Obj RefId="0"><MS><I32 N="RunspaceState">1</I32></MS></Obj>'

    ps_value = deserialize(element)
    assert isinstance(ps_value, PSObject)
    assert not isinstance(ps_value, psrp.RunspacePoolStateMsg)
    assert ps_value.RunspaceState == 1
    assert isinstance(ps_value.RunspaceState, PSInt)
    assert not hasattr(ps_value, "ExceptionAsErrorRecord")
    assert len(ps_value.PSObject.extended_properties) == 1


def test_user_event():
    ps_value = psrp.UserEvent(
        EventIdentifier=1,
        SourceIdentifier="source id",
        TimeGenerated=PSDateTime(1970, 1, 1),
        Sender="sender",
        SourceArgs="source args",
        MessageData="message data",
        ComputerName="computer name",
        RunspaceId=PSGuid("85ba13fb-6804-47f0-861a-8fe0ceb04acd"),
    )
    assert ps_value.EventIdentifier == 1
    assert ps_value.SourceIdentifier == "source id"
    assert ps_value.TimeGenerated == PSDateTime(1970, 1, 1)
    assert ps_value.Sender == "sender"
    assert ps_value.SourceArgs == "source args"
    assert ps_value.MessageData == "message data"
    assert ps_value.ComputerName == "computer name"
    assert ps_value.RunspaceId == PSGuid("85ba13fb-6804-47f0-861a-8fe0ceb04acd")

    element = serialize(ps_value)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert (
        actual == '<Obj RefId="0">'
        "<MS>"
        '<I32 N="PSEventArgs.EventIdentifier">1</I32>'
        '<S N="PSEventArgs.SourceIdentifier">source id</S>'
        '<DT N="PSEventArgs.TimeGenerated">1970-01-01T00:00:00Z</DT>'
        '<S N="PSEventArgs.Sender">sender</S>'
        '<S N="PSEventArgs.SourceArgs">source args</S>'
        '<S N="PSEventArgs.MessageData">message data</S>'
        '<S N="PSEventArgs.ComputerName">computer name</S>'
        '<G N="PSEventArgs.RunspaceId">85ba13fb-6804-47f0-861a-8fe0ceb04acd</G>'
        "</MS>"
        "</Obj>"
    )

    ps_value = deserialize(element)
    assert isinstance(ps_value, PSObject)
    assert not isinstance(ps_value, psrp.UserEvent)
    assert ps_value["PSEventArgs.EventIdentifier"] == 1
    assert ps_value["PSEventArgs.SourceIdentifier"] == "source id"
    assert ps_value["PSEventArgs.TimeGenerated"] == PSDateTime(1970, 1, 1, tzinfo=datetime.timezone.utc)
    assert ps_value["PSEventArgs.Sender"] == "sender"
    assert ps_value["PSEventArgs.SourceArgs"] == "source args"
    assert ps_value["PSEventArgs.MessageData"] == "message data"
    assert ps_value["PSEventArgs.ComputerName"] == "computer name"
    assert ps_value["PSEventArgs.RunspaceId"] == PSGuid("85ba13fb-6804-47f0-861a-8fe0ceb04acd")


def test_connect_runspace_pool():
    ps_value = psrp.ConnectRunspacePool(1, 2)

    element = serialize(ps_value)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    expected = '<Obj RefId="0"><MS><I32 N="MinRunspaces">1</I32><I32 N="MaxRunspaces">2</I32></MS></Obj>'
    assert actual == expected

    ps_value = deserialize(element)
    assert isinstance(ps_value, PSObject)
    assert not isinstance(ps_value, psrp.UserEvent)
    assert ps_value.MinRunspaces == 1
    assert ps_value.MaxRunspaces == 2


def test_connect_runspace_pool_empty():
    ps_value = psrp.ConnectRunspacePool()

    element = serialize(ps_value)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    expected = "<S />"
    assert actual == expected

    ps_value = deserialize(element)
    assert isinstance(ps_value, PSObject)
    assert not isinstance(ps_value, psrp.UserEvent)
    assert isinstance(ps_value, PSString)
    assert ps_value == ""


def test_error_record():
    ps_value = psrp.ErrorRecordMsg(
        Exception=NETException(Message="exception message"),
        CategoryInfo=ErrorCategoryInfo(),
    )

    element = serialize(ps_value)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    expected = (
        '<Obj RefId="0">'
        '<TN RefId="0"><T>System.Management.Automation.ErrorRecord</T><T>System.Object</T></TN>'
        "<MS>"
        '<Obj RefId="1" N="Exception">'
        '<TN RefId="1"><T>System.Exception</T><T>System.Object</T></TN>'
        '<Props><S N="Message">exception message</S>'
        '<Nil N="Data" />'
        '<Nil N="HelpLink" />'
        '<Nil N="HResult" />'
        '<Nil N="InnerException" />'
        '<Nil N="Source" />'
        '<Nil N="StackTrace" />'
        '<Nil N="TargetSite" />'
        "</Props>"
        "</Obj>"
        '<Nil N="TargetObject" />'
        '<Nil N="FullyQualifiedErrorId" />'
        '<Nil N="InvocationInfo" />'
        '<I32 N="ErrorCategory_Category">0</I32>'
        '<Nil N="ErrorCategory_Activity" />'
        '<Nil N="ErrorCategory_Reason" />'
        '<Nil N="ErrorCategory_TargetName" />'
        '<Nil N="ErrorCategory_TargetType" />'
        '<S N="ErrorCategory_Message">NotSpecified (:) [], </S>'
        '<B N="SerializeExtendedInfo">false</B>'
        "</MS>"
        "<ToString>exception message</ToString>"
        "</Obj>"
    )
    assert_xml_diff(actual, expected)

    ps_value = deserialize(element)
    assert isinstance(ps_value, PSObject)
    assert not isinstance(ps_value, psrp.ErrorRecordMsg)
    assert isinstance(ps_value, psrp.ErrorRecord)
    assert isinstance(ps_value.Exception, NETException)
    assert ps_value.Exception.Message == "exception message"
    assert ps_value.CategoryInfo.Category == ErrorCategory.NotSpecified
    assert ps_value.CategoryInfo.Activity is None
    assert ps_value.CategoryInfo.Reason is None
    assert ps_value.CategoryInfo.TargetName is None
    assert ps_value.CategoryInfo.TargetType is None
    assert ps_value.TargetObject is None
    assert ps_value.FullyQualifiedErrorId is None
    assert ps_value.InvocationInfo is None
    assert ps_value.ErrorDetails is None
    assert ps_value.PipelineIterationInfo is None
    assert ps_value.ScriptStackTrace is None


def test_debug_record():
    ps_value = psrp.DebugRecordMsg("msg")
    element = serialize(ps_value)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    expected = (
        '<Obj RefId="0">'
        '<TN RefId="0">'
        "<T>System.Management.Automation.DebugRecord</T>"
        "<T>System.Management.Automation.InformationalRecord</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<MS>"
        '<S N="InformationalRecord_Message">msg</S>'
        '<B N="InformationalRecord_SerializeInvocationInfo">false</B>'
        "</MS>"
        "</Obj>"
    )
    assert actual == expected

    ps_value = deserialize(element)

    assert isinstance(ps_value, PSObject)
    assert not isinstance(ps_value, psrp.DebugRecordMsg)
    assert isinstance(ps_value, DebugRecord)
    assert ps_value.Message == "msg"
    assert ps_value.InvocationInfo is None
    assert ps_value.PipelineIterationInfo is None


def test_information_record():
    ps_value = psrp.InformationRecordMsg(MessageData=1, Source="source")

    element = serialize(ps_value)
    actual = ElementTree.tostring(element, encoding="utf-8", method="xml").decode()
    expected = (
        '<Obj RefId="0">'
        "<MS>"
        '<I32 N="MessageData">1</I32>'
        '<S N="Source">source</S>'
        '<Nil N="TimeGenerated" />'
        '<Nil N="Tags" />'
        '<Nil N="User" />'
        '<Nil N="Computer" />'
        '<Nil N="ProcessId" />'
        '<Nil N="NativeThreadId" />'
        '<Nil N="ManagedThreadId" />'
        "</MS>"
        "</Obj>"
    )
    assert actual == expected

    ps_value = deserialize(element)
    assert isinstance(ps_value, PSObject)
    assert not isinstance(ps_value, psrp.InformationRecordMsg)
    assert isinstance(ps_value, PSObject)
    assert ps_value.MessageData == 1
    assert ps_value.Source == "source"
    assert ps_value.TimeGenerated is None
    assert ps_value.Tags is None
    assert ps_value.User is None
    assert ps_value.Computer is None
    assert ps_value.ProcessId is None
    assert ps_value.NativeThreadId is None
    assert ps_value.ManagedThreadId is None



# -*- coding: utf-8 -*-
# Copyright: (c) 2021, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import re
import uuid

import pytest

import psrpcore
from psrpcore._pipeline import PowerShell
from psrpcore.types import (
    ErrorCategoryInfo,
    ErrorRecord,
    HostMethodIdentifier,
    NETException,
    PSInvocationState,
    RunspacePoolState,
)

from .conftest import get_runspace_pair


def test_close_with_begin():
    client, server = get_runspace_pair()

    server.begin_close()
    assert client.state == RunspacePoolState.Opened
    assert server.state == RunspacePoolState.Closing

    client.receive_data(server.data_to_send())
    state = client.next_event()
    assert isinstance(state, psrpcore.RunspacePoolStateEvent)
    assert state.state == RunspacePoolState.Closing
    assert state.reason is None
    assert client.state == RunspacePoolState.Closing
    assert server.state == RunspacePoolState.Closing

    server.close()
    assert client.state == RunspacePoolState.Closing
    assert server.state == RunspacePoolState.Closed

    client.receive_data(server.data_to_send())
    state = client.next_event()
    assert isinstance(state, psrpcore.RunspacePoolStateEvent)
    assert state.state == RunspacePoolState.Closed
    assert state.reason is None
    assert client.state == RunspacePoolState.Closed
    assert server.state == RunspacePoolState.Closed


def test_server_is_broken():
    client, server = get_runspace_pair()

    err = ErrorRecord(
        Exception=NETException(Message="exception message"),
        CategoryInfo=ErrorCategoryInfo(),
    )
    server.set_broken(err)
    assert client.state == RunspacePoolState.Opened
    assert server.state == RunspacePoolState.Broken

    client.receive_data(server.data_to_send())
    state = client.next_event()
    assert isinstance(state, psrpcore.RunspacePoolStateEvent)
    assert state.state == RunspacePoolState.Broken
    assert isinstance(state.reason, ErrorRecord)
    assert str(state.reason) == "exception message"
    assert client.state == RunspacePoolState.Broken
    assert server.state == RunspacePoolState.Broken


def test_connect_already_connected():
    server = get_runspace_pair()[1]
    server.connect()
    assert server.data_to_send() is None


def test_connect_invalid_state():
    server = psrpcore.ServerRunspacePool()

    expected = re.escape(
        "Runspace Pool state must be one of 'RunspacePoolState.Disconnected' to accept Runspace Pool connections, "
        "current state is RunspacePoolState.BeforeOpen"
    )
    with pytest.raises(psrpcore.InvalidRunspacePoolState, match=expected):
        server.connect()


def test_send_event_invalid_state():
    server = psrpcore.ServerRunspacePool()

    expected = re.escape(
        "Runspace Pool state must be one of 'RunspacePoolState.Opened' to generate a Runspace Pool event, "
        "current state is RunspacePoolState.BeforeOpen"
    )
    with pytest.raises(psrpcore.InvalidRunspacePoolState, match=expected):
        server.send_event(1, "source")


def test_set_broken():
    client, server = get_runspace_pair()

    server.set_broken(ErrorRecord(NETException("error"), ErrorCategoryInfo()))
    assert client.state == RunspacePoolState.Opened
    assert server.state == RunspacePoolState.Broken

    client.receive_data(server.data_to_send())
    assert server.data_to_send() is None

    state = client.next_event()
    assert isinstance(state, psrpcore.RunspacePoolStateEvent)
    assert state.state == RunspacePoolState.Broken
    assert isinstance(state.reason, ErrorRecord)
    assert str(state.reason) == "error"
    assert client.state == RunspacePoolState.Broken


def test_set_broken_invalid_state():
    server = psrpcore.ServerRunspacePool()

    expected = re.escape(
        "Runspace Pool state must be one of 'RunspacePoolState.Broken, RunspacePoolState.Opened' to set as broken, "
        "current state is RunspacePoolState.BeforeOpen"
    )
    with pytest.raises(psrpcore.InvalidRunspacePoolState, match=expected):
        server.set_broken(ErrorRecord(NETException("error"), ErrorCategoryInfo()))


def test_host_call_invalid_state():
    server = psrpcore.ServerRunspacePool()

    expected = re.escape(
        "Runspace Pool state must be one of 'RunspacePoolState.Opened' to create host call, "
        "current state is RunspacePoolState.BeforeOpen"
    )
    with pytest.raises(psrpcore.InvalidRunspacePoolState, match=expected):
        server.host_call(HostMethodIdentifier.Write1, ["test"])


def test_request_key_invalid_state():
    server = psrpcore.ServerRunspacePool()

    expected = re.escape(
        "Runspace Pool state must be one of 'RunspacePoolState.Opened' to request exchange key, "
        "current state is RunspacePoolState.BeforeOpen"
    )
    with pytest.raises(psrpcore.InvalidRunspacePoolState, match=expected):
        server.request_key()


def test_connect_invalid_runspace():
    client, server = get_runspace_pair()
    client2 = psrpcore.ClientRunspacePool(runspace_pool_id=uuid.UUID(int=0))
    client.disconnect()
    server.disconnect()

    client2.connect()
    server.connect()
    server.receive_data(client2.data_to_send())

    with pytest.raises(
        psrpcore.PSRPCoreError, match="Incoming connection is targeted towards a different Runspace Pool"
    ):
        server.next_event()


def test_start_pipeline():
    client, server = get_runspace_pair()
    ps = psrpcore.ClientPowerShell(client)
    ps.add_script("test")
    ps.start()

    pipeline = psrpcore.ServerPipeline(server, ps.pipeline_id)
    server.receive_data(client.data_to_send())
    create_pipe = server.next_event()
    assert isinstance(create_pipe, psrpcore.CreatePipelineEvent)
    assert create_pipe.pipeline_id == ps.pipeline_id
    assert isinstance(create_pipe.pipeline, PowerShell)
    assert pipeline.state == PSInvocationState.NotStarted

    pipeline.start()
    assert pipeline.state == PSInvocationState.Running

    client.receive_data(server.data_to_send())
    state = client.next_event()
    assert isinstance(state, psrpcore.PipelineStateEvent)
    assert state.pipeline_id == ps.pipeline_id
    assert state.state == PSInvocationState.Running
    assert ps.state == PSInvocationState.Running

    expected = re.escape(
        "Pipeline state must be one of 'PSInvocationState.NotStarted, PSInvocationState.Stopped, "
        "PSInvocationState.Completed' to start a pipeline, current state is PSInvocationState.Running"
    )
    with pytest.raises(psrpcore.InvalidPipelineState, match=expected):
        pipeline.start()

    expected = re.escape(
        "Pipeline state must be one of 'PSInvocationState.NotStarted, PSInvocationState.Stopped, "
        "PSInvocationState.Stopping, PSInvocationState.Completed, PSInvocationState.Failed' to closing a pipeline, "
        "current state is PSInvocationState.Running"
    )
    with pytest.raises(psrpcore.InvalidPipelineState, match=expected):
        pipeline.close()

    pipeline.complete()
    assert pipeline.state == PSInvocationState.Completed

    client.receive_data(server.data_to_send())
    state = client.next_event()
    assert isinstance(state, psrpcore.PipelineStateEvent)
    assert state.pipeline_id == ps.pipeline_id
    assert state.state == PSInvocationState.Completed
    assert ps.state == PSInvocationState.Completed

    # Start the pipeline again
    ps.start()
    server.receive_data(client.data_to_send())
    create_pipe = server.next_event()
    assert isinstance(create_pipe, psrpcore.CreatePipelineEvent)
    assert create_pipe.pipeline_id == ps.pipeline_id
    assert isinstance(create_pipe.pipeline, PowerShell)
    assert pipeline.state == PSInvocationState.Completed

    pipeline.start()
    assert pipeline.state == PSInvocationState.Running

    pipeline.complete()
    pipeline.close()
    assert pipeline.pipeline_id not in server.pipeline_table

    # Verify it doesn't fail if closing an already closed pipeline
    pipeline.close()


def test_stop_pipeline():
    client, server = get_runspace_pair()
    ps = psrpcore.ClientPowerShell(client)
    ps.add_script("test")
    ps.start()

    pipeline = psrpcore.ServerPipeline(server, ps.pipeline_id)
    server.receive_data(client.data_to_send())
    create_pipe = server.next_event()
    assert isinstance(create_pipe, psrpcore.CreatePipelineEvent)
    assert create_pipe.pipeline_id == ps.pipeline_id
    assert isinstance(create_pipe.pipeline, PowerShell)
    assert pipeline.state == PSInvocationState.NotStarted

    expected = re.escape(
        "Pipeline state must be one of 'PSInvocationState.Running' to stop a pipeline, current state is "
        "PSInvocationState.NotStarted"
    )
    with pytest.raises(psrpcore.InvalidPipelineState, match=expected):
        pipeline.stop()

    pipeline.start()
    assert pipeline.state == PSInvocationState.Running
    client.receive_data(server.data_to_send())
    state = client.next_event()
    assert isinstance(state, psrpcore.PipelineStateEvent)
    assert state.pipeline_id == ps.pipeline_id
    assert state.state == PSInvocationState.Running
    assert ps.state == PSInvocationState.Running

    pipeline.stop()
    client.receive_data(server.data_to_send())
    pipe_state = client.next_event()
    assert isinstance(pipe_state, psrpcore.PipelineStateEvent)
    assert pipe_state.state == PSInvocationState.Stopped
    assert isinstance(pipe_state.reason, ErrorRecord)
    assert str(pipe_state.reason) == "The pipeline has been stopped."
    assert pipeline.state == PSInvocationState.Stopped
    assert ps.state == PSInvocationState.Stopped

    pipeline.stop()
    assert pipeline.state == PSInvocationState.Stopped
    assert server.data_to_send() is None

    # Run again from a stopped state
    ps.start()
    server.receive_data(client.data_to_send())
    create_pipe = server.next_event()
    assert isinstance(create_pipe, psrpcore.CreatePipelineEvent)
    assert create_pipe.pipeline_id == ps.pipeline_id
    assert isinstance(create_pipe.pipeline, PowerShell)
    assert pipeline.state == PSInvocationState.Stopped

    pipeline.start()
    client.receive_data(server.data_to_send())
    state = client.next_event()
    assert isinstance(state, psrpcore.PipelineStateEvent)
    assert state.state == PSInvocationState.Running
    assert ps.state == PSInvocationState.Running

    pipeline.stop()
    client.receive_data(server.data_to_send())
    state = client.next_event()
    assert isinstance(state, psrpcore.PipelineStateEvent)
    assert state.state == PSInvocationState.Stopped
    assert ps.state == PSInvocationState.Stopped

    pipeline.close()
    assert pipeline.state == PSInvocationState.Stopped
    assert server.data_to_send() is None


def test_pipeline_host_call_invalid_state():
    client, server = get_runspace_pair()
    ps = psrpcore.ClientPowerShell(client)
    ps.add_script("test")
    ps.start()

    pipeline = psrpcore.ServerPipeline(server, ps.pipeline_id)
    server.receive_data(client.data_to_send())
    server.next_event()

    expected = re.escape(
        "Pipeline state must be one of 'PSInvocationState.Running' to make a pipeline host call, current state is "
        "PSInvocationState.NotStarted"
    )
    with pytest.raises(psrpcore.InvalidPipelineState, match=expected):
        pipeline.host_call(HostMethodIdentifier.Write1, ["line"])


def test_pipeline_write_output_invalid_state():
    client, server = get_runspace_pair()
    ps = psrpcore.ClientPowerShell(client)
    ps.add_script("test")
    ps.start()

    pipeline = psrpcore.ServerPipeline(server, ps.pipeline_id)
    server.receive_data(client.data_to_send())
    server.next_event()

    expected = re.escape(
        "Pipeline state must be one of 'PSInvocationState.Running' to write pipeline output, current state is "
        "PSInvocationState.NotStarted"
    )
    with pytest.raises(psrpcore.InvalidPipelineState, match=expected):
        pipeline.write_output("value")


def test_pipeline_write_error_invalid_state():
    client, server = get_runspace_pair()
    ps = psrpcore.ClientPowerShell(client)
    ps.add_script("test")
    ps.start()

    pipeline = psrpcore.ServerPipeline(server, ps.pipeline_id)
    server.receive_data(client.data_to_send())
    server.next_event()

    expected = re.escape(
        "Pipeline state must be one of 'PSInvocationState.Running' to write pipeline error, current state is "
        "PSInvocationState.NotStarted"
    )
    with pytest.raises(psrpcore.InvalidPipelineState, match=expected):
        pipeline.write_error(NETException("error"))


def test_pipeline_write_debug_invalid_state():
    client, server = get_runspace_pair()
    ps = psrpcore.ClientPowerShell(client)
    ps.add_script("test")
    ps.start()

    pipeline = psrpcore.ServerPipeline(server, ps.pipeline_id)
    server.receive_data(client.data_to_send())
    server.next_event()

    expected = re.escape(
        "Pipeline state must be one of 'PSInvocationState.Running' to write pipeline debug, current state is "
        "PSInvocationState.NotStarted"
    )
    with pytest.raises(psrpcore.InvalidPipelineState, match=expected):
        pipeline.write_debug("value")


def test_pipeline_write_verbose_invalid_state():
    client, server = get_runspace_pair()
    ps = psrpcore.ClientPowerShell(client)
    ps.add_script("test")
    ps.start()

    pipeline = psrpcore.ServerPipeline(server, ps.pipeline_id)
    server.receive_data(client.data_to_send())
    server.next_event()

    expected = re.escape(
        "Pipeline state must be one of 'PSInvocationState.Running' to write pipeline verbose, current state is "
        "PSInvocationState.NotStarted"
    )
    with pytest.raises(psrpcore.InvalidPipelineState, match=expected):
        pipeline.write_verbose("value")


def test_pipeline_write_warning_invalid_state():
    client, server = get_runspace_pair()
    ps = psrpcore.ClientPowerShell(client)
    ps.add_script("test")
    ps.start()

    pipeline = psrpcore.ServerPipeline(server, ps.pipeline_id)
    server.receive_data(client.data_to_send())
    server.next_event()

    expected = re.escape(
        "Pipeline state must be one of 'PSInvocationState.Running' to write pipeline warning, current state is "
        "PSInvocationState.NotStarted"
    )
    with pytest.raises(psrpcore.InvalidPipelineState, match=expected):
        pipeline.write_warning("value")


def test_pipeline_write_progress_invalid_state():
    client, server = get_runspace_pair()
    ps = psrpcore.ClientPowerShell(client)
    ps.add_script("test")
    ps.start()

    pipeline = psrpcore.ServerPipeline(server, ps.pipeline_id)
    server.receive_data(client.data_to_send())
    server.next_event()

    expected = re.escape(
        "Pipeline state must be one of 'PSInvocationState.Running' to write pipeline progress, current state is "
        "PSInvocationState.NotStarted"
    )
    with pytest.raises(psrpcore.InvalidPipelineState, match=expected):
        pipeline.write_progress("activity", 1, "status")


def test_pipeline_information_invalid_protocol():
    client, server = get_runspace_pair()
    ps = psrpcore.ClientPowerShell(client)
    ps.add_script("test")
    ps.start()

    pipeline = psrpcore.ServerPipeline(server, ps.pipeline_id)
    server.receive_data(client.data_to_send())
    server.next_event()

    server.their_capability = None
    expected = re.escape("writing information record requires a protocol version of 2.3, current version is 2.0")
    with pytest.raises(psrpcore.InvalidProtocolVersion, match=expected):
        pipeline.write_information("message", "source")


def test_pipeline_information_invalid_state():
    client, server = get_runspace_pair()
    ps = psrpcore.ClientPowerShell(client)
    ps.add_script("test")
    ps.start()

    pipeline = psrpcore.ServerPipeline(server, ps.pipeline_id)
    server.receive_data(client.data_to_send())
    server.next_event()

    expected = re.escape(
        "Pipeline state must be one of 'PSInvocationState.Running' to write pipeline information, current state is "
        "PSInvocationState.NotStarted"
    )
    with pytest.raises(psrpcore.InvalidPipelineState, match=expected):
        pipeline.write_information("message", "source")


def test_pipeline_host_call():
    client, server = get_runspace_pair()

    c_pipeline = psrpcore.ClientPowerShell(client)
    # This is meant to be parsed by some engine and isn't actually read in this test
    # Just used to indicate a scenario where this would occur.
    c_pipeline.add_script('$host.UI.PromptForCredential("caption", "message", "username", "targetname")')
    c_pipeline.start()

    s_pipeline = psrpcore.ServerPipeline(server, c_pipeline.pipeline_id)
    server.receive_data(client.data_to_send())
    server.next_event()
    s_pipeline.start()

    client.receive_data(server.data_to_send())
    client.next_event()

    s_host = psrpcore.ServerHostRequestor(s_pipeline)
    s_host.prompt_for_credential("caption", "message", "username", "targetname")

    client.receive_data(server.data_to_send())
    host_call = client.next_event()
    assert isinstance(host_call, psrpcore.PipelineHostCallEvent)
    assert host_call.ci == 1
    assert host_call.method_identifier == HostMethodIdentifier.PromptForCredential1
    assert host_call.method_parameters == ["caption", "message", "username", "targetname"]

    c_host = psrpcore.ClientHostResponder(c_pipeline)
    c_host.prompt_for_credential(1, "prompt response")
    server.receive_data(client.data_to_send())
    host_response = server.next_event()
    assert isinstance(host_response, psrpcore.PipelineHostResponseEvent)
    assert host_response.ci == 1
    assert host_response.method_identifier == HostMethodIdentifier.PromptForCredential1
    assert host_response.result == "prompt response"



# -*- coding: utf-8 -*-
# Copyright: (c) 2021, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

"""PSRP/.NET Enum Types.

The base classes for any .NET enum type.
"""

import enum
import typing

from psrpcore.types._base import PSType
from psrpcore.types._primitive import PSInt, PSIntegerBase


class PSEnumMeta(enum.EnumMeta):
    """The meta type for all PowerShell enum objects.

    This is the meta type that extends the enum meta type to support .NET enum
    specific info. Any .NET enum type can specify ``base_type`` as a PS integer
    type to change the fundamental integer type each enum value is based on. By
    default it is :class:`PSInt`.

    This class is used internally and is not designed for public consumption.
    You should inherit from the existing base classes that have already set
    this as their metaclass.
    """

    @classmethod
    def __prepare__(  # type: ignore[override]  # Uses internal class for annotation
        metacls,
        __name: str,
        __bases: typing.Tuple[type, ...],
        **kwds: typing.Any,
    ) -> typing.Mapping[str, typing.Any]:
        # Python <3.9 will fail when passing the kwds if base_type was specified, it's omitted entirely here.
        return super().__prepare__(__name, __bases)

    def __new__(
        mcls,
        name: str,
        bases: typing.Tuple[type, ...],
        namespace: typing.Dict[str, typing.Any],
        **kw: typing.Type[PSIntegerBase],
    ) -> "PSEnumMeta":
        # Ensure the enum values are casted to the PS integer type for serialization.
        base_type = kw.get("base_type", PSInt)
        if not isinstance(base_type, type) or not issubclass(base_type, PSIntegerBase):
            raise TypeError(f"PSEnumType {name} base_type must be a subclass of PSIntegerBase")

        def new(cls: typing.Type, val: typing.Any) -> typing.Type:
            val = base_type(val)
            obj = int.__new__(cls, val)
            obj._value_ = val

            return obj

        namespace["__new__"] = new

        return super().__new__(
            mcls,
            name,
            bases,
            namespace,  # type: ignore[arg-type] # _EnumDict is private so cannot be used properly
        )


@PSType(["System.Enum", "System.ValueType"], rehydrate=False)
class PSEnumBase(PSIntegerBase, enum.Enum, metaclass=PSEnumMeta):
    """The base enum PSObject type.

    This is the base enum PSObject type that all enum complex objects should
    inherit from. Any objects that inherit `PSEnumBase` and require a base type
    that is not :class:`PSInt` should set `base_type=...` when declaring the
    class.

    An example enum would look like:

    .. code-block:: python

        @PSType(["System.MyEnum"])
        class MyEnum(PSEnumBase):
            Label = 1
            Other = 2

        @PSType(["System.MyUIntEnum"])
        class MyUIntEnum(PSEnumBase, base_type=PSUInt):
            Label = 1
            Other = 0xFFFFFFFF

    A user of that enum would then access it like `MyEnum.Label` or
    `MyEnum.Other`. This class is designed for enums that allow only 1 value,
    if you require a flag like enum, use :class:`PSFlagBase` as the base type.
    """

    def __repr__(self) -> str:
        return enum.Enum.__repr__(self)

    def __str__(self) -> str:
        return enum.Enum.__str__(self)


@PSType(["System.Enum", "System.ValueType"], rehydrate=False)
class PSFlagBase(PSIntegerBase, enum.Flag, metaclass=PSEnumMeta):
    """The base flags enum PSObject type.

    This is like :class:`PSEnumBase` but supports having multiple values set
    like `[Flags]` on an enum in .NET. Using any bitwise operations will
    preserve the type so `MyFlags.Flag1 | MyFlags.Flag2` will still be an
    instance of `MyFlags`.

    Like :class:`PSEnumBase`, an implementing type can set `base_type` to
    another PS integer type if the base integer type is not Int32. An example
    flag enum would look like:

    .. code-block:: python

        @PSType(["System.MyFlags"])
        class MyFlags(PSFlagBase):
            Flag1 = 1
            Flag2 = 2
            Flag3 = 4

        @PSType(["System.MyUIntFlags"])
        class MyUIntFlags(PSFlagBase, base_type=PSUInt):
            Flag1 = 1
            Flag2 = 2
            Flag3 = 4
            All = 0xFFFFFFFF
    """

    # We ignore most of these mypy errors due to the weird __mro__ setup

    @classmethod
    def _missing_(cls, value):  # type: ignore[no-untyped-def]
        # Calls the unbound func so it runs the operations against our class.
        return enum.IntFlag._missing_.__func__(cls, value)  # type: ignore[attr-defined]

    @classmethod
    def _create_pseudo_member_(cls, value):  # type: ignore[no-untyped-def]
        # Calls the unbound func so it runs the operations against our class.
        return enum.IntFlag._create_pseudo_member_.__func__(cls, value)  # type: ignore[attr-defined]

    def __or__(self, other):  # type: ignore[no-untyped-def]
        return enum.IntFlag.__or__(self, other)

    def __and__(self, other):  # type: ignore[no-untyped-def]
        return enum.IntFlag.__and__(self, other)

    def __xor__(self, other):  # type: ignore[no-untyped-def]
        return enum.IntFlag.__xor__(self, other)

    def __ror__(self, other):  # type: ignore[no-untyped-def]
        return enum.IntFlag.__or__(self, other)

    def __rand__(self, other):  # type: ignore[no-untyped-def]
        return enum.IntFlag.__and__(self, other)

    def __rxor__(self, other):  # type: ignore[no-untyped-def]
        return enum.IntFlag.__xor__(self, other)

    def __invert__(self):  # type: ignore[no-untyped-def]
        return enum.IntFlag.__invert__(self)

    def __repr__(self) -> str:
        return enum.IntFlag.__repr__(self)

    def __str__(self) -> str:
        return enum.IntFlag.__str__(self)



# -*- coding: utf-8 -*-
# Copyright: (c) 2021, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import collections
import queue
import re
import xml.etree.ElementTree as ElementTree

import pytest

import psrpcore.types._collection as collection
from psrpcore.types import PSChar, PSInt64, PSNoteProperty

from ..conftest import (
    COMPLEX_ENCODED_STRING,
    COMPLEX_STRING,
    assert_xml_diff,
    deserialize,
    serialize,
)


def test_ps_dict_instantiation():
    expected = re.escape(
        "Type PSDictBase cannot be instantiated; it can be used only as a base class for dictionary " "types."
    )
    with pytest.raises(TypeError, match=expected):
        collection.PSDictBase()


def test_ps_list_instantiation():
    expected = re.escape(
        "Type PSListBase cannot be instantiated; it can be used only as a base class for list " "types."
    )
    with pytest.raises(TypeError, match=expected):
        collection.PSListBase()


def test_ps_queue_instantiation():
    expected = re.escape(
        "Type PSQueueBase cannot be instantiated; it can be used only as a base class for queue " "types."
    )
    with pytest.raises(TypeError, match=expected):
        collection.PSQueueBase()


def test_ps_stack_instantiation():
    expected = re.escape(
        "Type PSStackBase cannot be instantiated; it can be used only as a base class for list " "types."
    )
    with pytest.raises(TypeError, match=expected):
        collection.PSStackBase()


def test_ps_stack():
    ps_value = collection.PSStack(["abc", 123, PSInt64(1)])
    ps_value.append(True)
    assert isinstance(ps_value, collection.PSStack)
    assert isinstance(ps_value, list)

    element = serialize(ps_value)
    actual = ElementTree.tostring(element, encoding="utf-8", method="xml").decode()
    assert (
        actual == '<Obj RefId="0">'
        '<TN RefId="0"><T>System.Collections.Stack</T><T>System.Object</T></TN>'
        "<STK><S>abc</S><I32>123</I32><I64>1</I64><B>true</B></STK>"
        "</Obj>"
    )

    actual = deserialize(element)
    assert isinstance(actual, collection.PSStack)
    assert isinstance(actual, list)
    assert str(actual) == "['abc', 123, 1, True]"
    assert repr(actual) == "['abc', 123, 1, True]"
    assert actual == ["abc", 123, PSInt64(1), True]
    # Verify we can still index the list
    assert actual[0] == "abc"
    assert actual[1] == 123
    assert actual[2] == PSInt64(1)
    assert actual[3] is True
    assert actual.PSTypeNames == ["System.Collections.Stack", "System.Object"]


def test_ps_stack_with_properties():
    ps_value = collection.PSStack([0, 2, PSChar("a")])
    ps_value.PSObject.extended_properties.append(PSNoteProperty("1"))
    ps_value[1] = 1
    ps_value["1"] = collection.PSStack(["123", 123])

    # Make sure we can access the stack using an index and the properties with a string.
    assert ps_value[1] == 1
    assert isinstance(ps_value["1"], collection.PSStack)
    assert ps_value["1"] == ["123", 123]

    # Check that appending an item doesn't clear our properties
    ps_value.append(2)

    element = serialize(ps_value)
    actual = ElementTree.tostring(element, encoding="utf-8", method="xml").decode()
    expected = (
        '<Obj RefId="0"><TN RefId="0"><T>System.Collections.Stack</T><T>System.Object</T></TN>'
        "<MS>"
        '<Obj RefId="1" N="1"><TNRef RefId="0" />'
        "<STK><S>123</S><I32>123</I32></STK>"
        "</Obj>"
        "</MS>"
        "<STK><I32>0</I32><I32>1</I32><C>97</C><I32>2</I32></STK>"
        "</Obj>"
    )
    assert_xml_diff(actual, expected)

    actual = deserialize(element)
    assert isinstance(actual, collection.PSStack)
    assert isinstance(actual, list)
    assert str(actual) == "[0, 1, 97, 2]"
    assert repr(actual) == "[0, 1, 97, 2]"
    assert actual == [0, 1, PSChar("a"), 2]
    # Verify we can still index the list
    assert actual[0] == 0
    assert actual[1] == 1
    assert actual[2] == PSChar("a")
    assert actual[3] == 2

    # Verify we can access the extended prop using a string index.
    assert isinstance(actual["1"], collection.PSStack)
    assert actual["1"] == collection.PSStack(["123", 123])

    assert actual.PSTypeNames == ["System.Collections.Stack", "System.Object"]


def test_ps_queue():
    ps_value = collection.PSQueue()
    ps_value.put("abc")
    ps_value.put(123)
    ps_value.put(PSInt64(1))
    ps_value.put(collection.PSQueue())
    assert isinstance(ps_value, collection.PSQueue)
    assert isinstance(ps_value, queue.Queue)

    element = serialize(ps_value)
    actual = ElementTree.tostring(element, encoding="utf-8", method="xml").decode()
    assert (
        actual == '<Obj RefId="0"><TN RefId="0"><T>System.Collections.Queue</T><T>System.Object</T></TN>'
        "<QUE>"
        "<S>abc</S>"
        "<I32>123</I32>"
        "<I64>1</I64>"
        '<Obj RefId="1"><TNRef RefId="0" /><QUE /></Obj>'
        "</QUE>"
        "</Obj>"
    )

    actual = deserialize(element)
    assert isinstance(actual, collection.PSQueue)
    assert isinstance(actual, queue.Queue)
    assert str(actual) == queue.Queue.__str__(actual)
    assert repr(actual) == queue.Queue.__repr__(actual)

    assert actual.get() == "abc"
    assert actual.get() == 123
    assert actual.get() == PSInt64(1)

    queue_entry = actual.get()
    assert isinstance(queue_entry, collection.PSQueue)
    assert isinstance(queue_entry, queue.Queue)
    with pytest.raises(queue.Empty):
        queue_entry.get(block=False)

    with pytest.raises(queue.Empty):
        actual.get(block=False)

    assert actual.PSTypeNames == ["System.Collections.Queue", "System.Object"]


def test_ps_queue_from_queue():
    q = queue.Queue()
    q.put(1)
    q.put("1")
    q.put("a")

    element = serialize(q)
    actual = ElementTree.tostring(element, encoding="utf-8", method="xml").decode()
    assert (
        actual == '<Obj RefId="0">'
        '<TN RefId="0"><T>System.Collections.Queue</T><T>System.Object</T></TN>'
        "<QUE><I32>1</I32><S>1</S><S>a</S></QUE>"
        "</Obj>"
    )


def test_ps_queue_with_properties():
    ps_value = collection.PSQueue()
    ps_value.put("abc")
    ps_value.put(123)
    ps_value.put(PSInt64(1))
    ps_value.put(collection.PSQueue())

    ps_value.PSObject.extended_properties.append(PSNoteProperty("1"))
    ps_value["1"] = collection.PSQueue()
    ps_value["1"].put("entry")

    element = serialize(ps_value)
    actual = ElementTree.tostring(element, encoding="utf-8", method="xml").decode()
    expected = (
        '<Obj RefId="0"><TN RefId="0"><T>System.Collections.Queue</T><T>System.Object</T></TN>'
        "<MS>"
        '<Obj RefId="1" N="1"><TNRef RefId="0" /><QUE><S>entry</S></QUE></Obj>'
        "</MS>"
        "<QUE>"
        "<S>abc</S>"
        "<I32>123</I32>"
        "<I64>1</I64>"
        '<Obj RefId="2"><TNRef RefId="0" /><QUE /></Obj>'
        "</QUE>"
        "</Obj>"
    )
    assert_xml_diff(actual, expected)

    actual = deserialize(element)
    assert isinstance(actual, collection.PSQueue)
    assert isinstance(actual, queue.Queue)
    assert str(actual) == queue.Queue.__str__(actual)
    assert repr(actual) == queue.Queue.__repr__(actual)

    assert actual.get() == "abc"
    assert actual.get() == 123
    assert actual.get() == PSInt64(1)

    queue_entry = actual.get()
    assert isinstance(queue_entry, collection.PSQueue)
    assert isinstance(queue_entry, queue.Queue)
    with pytest.raises(queue.Empty):
        queue_entry.get(block=False)

    with pytest.raises(queue.Empty):
        actual.get(block=False)

    prop_queue = actual["1"]
    assert isinstance(prop_queue, collection.PSQueue)
    assert isinstance(prop_queue, queue.Queue)
    assert prop_queue.get() == "entry"
    with pytest.raises(queue.Empty):
        prop_queue.get(block=False)

    assert actual.PSTypeNames == ["System.Collections.Queue", "System.Object"]


def test_ps_list():
    ps_value = collection.PSList(["abc", 123, PSInt64(1)])
    ps_value.append(True)
    assert isinstance(ps_value, collection.PSList)
    assert isinstance(ps_value, list)

    element = serialize(ps_value)
    actual = ElementTree.tostring(element, encoding="utf-8", method="xml").decode()
    assert (
        actual == '<Obj RefId="0">'
        '<TN RefId="0"><T>System.Collections.ArrayList</T><T>System.Object</T></TN>'
        "<LST><S>abc</S><I32>123</I32><I64>1</I64><B>true</B></LST>"
        "</Obj>"
    )

    actual = deserialize(element)
    assert isinstance(actual, collection.PSList)
    assert isinstance(actual, list)
    assert str(actual) == "['abc', 123, 1, True]"
    assert repr(actual) == "['abc', 123, 1, True]"

    assert actual == ["abc", 123, PSInt64(1), True]
    # Verify we can still index the list
    assert actual[0] == "abc"
    assert actual[1] == 123
    assert actual[2] == PSInt64(1)
    assert actual[3] is True
    assert actual.PSTypeNames == ["System.Collections.ArrayList", "System.Object"]


def test_ps_list_from_list():
    element = serialize([1, "1", "a"])
    actual = ElementTree.tostring(element, encoding="utf-8", method="xml").decode()
    assert (
        actual == '<Obj RefId="0">'
        '<TN RefId="0"><T>System.Collections.ArrayList</T><T>System.Object</T></TN>'
        "<LST><I32>1</I32><S>1</S><S>a</S></LST>"
        "</Obj>"
    )


def test_ps_list_with_properties():
    ps_value = collection.PSList([0, 2, PSChar("a")])
    ps_value.PSObject.extended_properties.append(PSNoteProperty("1"))
    ps_value[1] = 1
    ps_value["1"] = collection.PSList(["123", 123])

    # Make sure we can access the stack using an index and the properties with a string.
    assert ps_value[1] == 1
    assert isinstance(ps_value["1"], collection.PSList)
    assert ps_value["1"] == ["123", 123]

    # Check that appending an item doesn't clear our properties
    ps_value.append(2)

    element = serialize(ps_value)
    actual = ElementTree.tostring(element, encoding="utf-8", method="xml").decode()
    expected = (
        '<Obj RefId="0"><TN RefId="0"><T>System.Collections.ArrayList</T><T>System.Object</T></TN>'
        "<MS>"
        '<Obj RefId="1" N="1"><TNRef RefId="0" />'
        "<LST><S>123</S><I32>123</I32></LST>"
        "</Obj>"
        "</MS>"
        "<LST><I32>0</I32><I32>1</I32><C>97</C><I32>2</I32></LST>"
        "</Obj>"
    )
    assert_xml_diff(actual, expected)

    actual = deserialize(element)
    assert isinstance(actual, collection.PSList)
    assert isinstance(actual, list)
    assert actual == [0, 1, PSChar("a"), 2]
    # Verify we can still index the list
    assert actual[0] == 0
    assert actual[1] == 1
    assert actual[2] == PSChar("a")
    assert actual[3] == 2

    # Verify we can access the extended prop using a string index.
    assert isinstance(actual["1"], collection.PSList)
    assert actual["1"] == collection.PSList(["123", 123])

    assert actual.PSTypeNames == ["System.Collections.ArrayList", "System.Object"]


@pytest.mark.parametrize(
    "input_value, expected",
    [
        ({}, "<DCT />"),
        ({"a": "a"}, '<DCT><En><S N="Key">a</S><S N="Value">a</S></En></DCT>'),
        ({"a": 1}, '<DCT><En><S N="Key">a</S><I32 N="Value">1</I32></En></DCT>'),
        (
            {1: PSChar("a"), PSInt64(10): ["abc", 456]},
            '<DCT><En><I32 N="Key">1</I32><C N="Value">97</C></En>'
            '<En><I64 N="Key">10</I64><Obj RefId="1" N="Value">'
            '<TN RefId="1"><T>System.Collections.ArrayList</T><T>System.Object</T></TN>'
            "<LST><S>abc</S><I32>456</I32></LST></Obj></En></DCT>",
        ),
    ],
)
def test_ps_dict(input_value, expected):
    ps_value = collection.PSDict(input_value)
    assert isinstance(ps_value, collection.PSDict)
    assert isinstance(ps_value, dict)

    element = serialize(ps_value)
    actual = ElementTree.tostring(element, encoding="utf-8", method="xml").decode()
    expected = (
        f'<Obj RefId="0"><TN RefId="0"><T>System.Collections.Hashtable</T><T>System.Object</T></TN>'
        f"{expected}"
        f"</Obj>"
    )
    assert_xml_diff(actual, expected)

    actual = deserialize(element)
    assert isinstance(actual, collection.PSDict)
    assert isinstance(actual, dict)
    assert actual == input_value
    assert actual.PSTypeNames == ["System.Collections.Hashtable", "System.Object"]


def test_ps_dict_from_dict():
    element = serialize({"abc": "def", 1: 2, PSChar("a"): collection.PSList([1, 2])})
    actual = ElementTree.tostring(element, encoding="utf-8", method="xml").decode()
    expected = (
        '<Obj RefId="0"><TN RefId="0"><T>System.Collections.Hashtable</T><T>System.Object</T></TN>'
        "<DCT>"
        "<En>"
        '<S N="Key">abc</S>'
        '<S N="Value">def</S>'
        "</En>"
        "<En>"
        '<I32 N="Key">1</I32>'
        '<I32 N="Value">2</I32>'
        "</En>"
        "<En>"
        '<C N="Key">97</C>'
        '<Obj RefId="1" N="Value">'
        '<TN RefId="1"><T>System.Collections.ArrayList</T><T>System.Object</T></TN>'
        "<LST><I32>1</I32><I32>2</I32></LST>"
        "</Obj>"
        "</En>"
        "</DCT>"
        "</Obj>"
    )
    assert_xml_diff(actual, expected)


def test_ps_dict_with_properties():
    ps_value = collection.PSDict({})
    ps_value.PSObject.extended_properties.append(PSNoteProperty("key"))

    complex_prop = PSNoteProperty(COMPLEX_STRING)
    ps_value.PSObject.extended_properties.append(complex_prop)

    other_prop = PSNoteProperty("other", "prop")
    ps_value.PSObject.extended_properties.append(other_prop)

    # Setting a value will always set it in the dict, even adding a new dict entry if the prop exists
    ps_value["key"] = "dict"
    ps_value[COMPLEX_STRING] = "dict"

    # We can still set a property using dot notation like in PowerShell
    ps_value.key = "prop"

    # Or on the property object itself if we cannot access it like a Python attribute
    complex_prop.set_value("prop", ps_value)

    element = serialize(ps_value)
    actual = ElementTree.tostring(element, encoding="utf-8", method="xml").decode()
    assert (
        actual == f'<Obj RefId="0"><TN RefId="0"><T>System.Collections.Hashtable</T><T>System.Object</T></TN>'
        f"<MS>"
        f'<S N="key">prop</S>'
        f'<S N="{COMPLEX_ENCODED_STRING}">prop</S>'
        f'<S N="other">prop</S>'
        f"</MS>"
        f"<DCT>"
        f'<En><S N="Key">key</S><S N="Value">dict</S></En>'
        f'<En><S N="Key">{COMPLEX_ENCODED_STRING}</S><S N="Value">dict</S></En>'
        f"</DCT>"
        f"</Obj>"
    )

    actual = deserialize(element)
    assert isinstance(actual, collection.PSDict)
    assert isinstance(actual, dict)
    assert str(actual) == f"{{'key': 'dict', {COMPLEX_STRING!r}: 'dict'}}"
    assert repr(actual) == f"{{'key': 'dict', {COMPLEX_STRING!r}: 'dict'}}"

    # In the case of a prop shadowing a dict, [] will favour the dict, and . will only get props
    assert actual["key"] == "dict"
    assert actual.key == "prop"

    # If only the prop exists under that name both [] and . will work
    assert actual["other"] == "prop"
    assert actual.other == "prop"

    # Because we cannot use special characters using the dot notation, we can only get shadowed props using the raw
    # PSObject property list
    assert actual.PSObject.extended_properties[1].name == COMPLEX_STRING
    assert actual.PSObject.extended_properties[1].get_value(actual) == "prop"


def test_ps_ienumerable():
    ps_value = collection.PSIEnumerable([0, 1, 2, 3, 4])
    assert isinstance(ps_value, collection.PSIEnumerable)
    assert isinstance(ps_value, list)

    element = serialize(ps_value)
    actual = ElementTree.tostring(element, encoding="utf-8", method="xml").decode()
    expected = (
        '<Obj RefId="0">'
        '<TN RefId="0">'
        "<T>System.Collections.IEnumerable</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<IE>"
        "<I32>0</I32>"
        "<I32>1</I32>"
        "<I32>2</I32>"
        "<I32>3</I32>"
        "<I32>4</I32>"
        "</IE>"
        "</Obj>"
    )
    assert actual == expected

    actual = deserialize(element)
    assert isinstance(actual, collection.PSIEnumerable)
    assert isinstance(actual, list)
    assert str(actual) == "[0, 1, 2, 3, 4]"
    assert repr(actual) == "[0, 1, 2, 3, 4]"

    assert actual == [0, 1, 2, 3, 4]
    assert actual.PSTypeNames == ["System.Collections.IEnumerable", "System.Object"]
